# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

__all__ = [
    "VolttronHomeFileReloader",
    "AbsolutePathFileReloader",
    "watch_file",
    "watch_file_with_fullpath",
]

import logging
import os

from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

from volttron.utils import ClientContext as cc

_log = logging.getLogger(__name__)


class VolttronHomeFileReloader(PatternMatchingEventHandler):
    """
    Extends PatternMatchingEvent handler to watch changes to a singlefile/file pattern within volttron home.
    filetowatch should be path relative to volttron home.
    For example filetowatch auth.json with watch file <volttron_home>/auth.json.
    filetowatch *.json will watch all json files in <volttron_home>
    """

    def __init__(self, filetowatch, callback):
        super(VolttronHomeFileReloader, self).__init__(patterns=[f"{cc.get_volttron_home()}/{filetowatch}"])
        _log.debug(f"patterns is {cc.get_volttron_home()}/{filetowatch}")
        self._callback = callback

    def on_closed(self, event):
        _log.debug("Calling callback on event {}. Calling {}".format(event, self._callback))
        try:
            self._callback()
        except BaseException as e:
            _log.error("Exception in callback: {}".format(e))
        _log.debug("After callback on event {}".format(event))


class AbsolutePathFileReloader(PatternMatchingEventHandler):
    """
    Extends PatternMatchingEvent handler to watch changes to a singlefile/file pattern within volttron home.
    filetowatch should be path relative to volttron home.
    For example filetowatch auth.json with watch file <volttron_home>/auth.json.
    filetowatch *.json will watch all json files in <volttron_home>
    """

    def __init__(self, filetowatch, callback):
        super(AbsolutePathFileReloader, self).__init__(patterns=[filetowatch])
        self._callback = callback
        self._filetowatch = filetowatch

    @property
    def watchfile(self):
        return self._filetowatch

    def on_closed(self, event):
        _log.debug("Calling callback on event {}. Calling {}".format(event, self._callback))
        try:
            self._callback(self._filetowatch)
        except BaseException as e:
            _log.error("Exception in callback: {}".format(e))
        _log.debug("After callback on event {}".format(event))


def watch_file(fullpath, callback):
    """Run callback method whenever the file changes

    Not available on OS X/MacOS.
    """

    dirname, filename = os.path.split(fullpath)
    _log.info(
        "Adding file watch for %s dirname=%s, filename=%s",
        fullpath,
        cc.get_volttron_home(),
        filename,
    )
    observer = Observer()
    observer.schedule(
        VolttronHomeFileReloader(filename, callback),
        path=cc.get_volttron_home(),
    )
    observer.start()
    _log.info("Added file watch for %s", fullpath)


def watch_file_with_fullpath(fullpath, callback):
    """Run callback method whenever the file changes

    Not available on OS X/MacOS.
    """
    dirname, filename = os.path.split(fullpath)
    _log.info("Adding file watch for %s", fullpath)
    _observer = Observer()
    _observer.schedule(AbsolutePathFileReloader(fullpath, callback), dirname)
    _log.info("Added file watch for %s", fullpath)
    _observer.start()



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging

from volttron.utils import get_address
from volttron.utils.keystore import KeyStore, KnownHostsStore
from volttron.client.vip.agent.connection import Connection

from volttron.client.logs import get_logger

_log = get_logger()

host_store = KnownHostsStore()


def get_known_host_serverkey(vip_address):
    return host_store.serverkey(vip_address)


def build_connection(identity,
                     peer="",
                     address=None,
                     publickey=None,
                     secretkey=None,
                     message_bus=None,
                     **kwargs):
    address = address if address is not None else get_address()
    if publickey is None or secretkey is None:
        publickey, secretkey = get_server_keys(publickey, secretkey)
    cn = Connection(address=address,
                    identity=identity,
                    peer=peer,
                    publickey=publickey,
                    secretkey=secretkey,
                    message_bus=message_bus,
                    **kwargs)
    return cn



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
""" Core package."""
from gevent import monkey

monkey.patch_all()

from volttron.utils import get_version

__version__ = get_version()

# def get_volttron_root():
#     """
#     Returns the root folder where the volttron code base resideds on disk.

    #     :return: absolute path to root folder
    #     """
    #     return os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

    # def get_volttron_data():
    #     root = get_volttron_root()
    #     return os.path.join(root, "volttron_data")

    # def get_services_core(agent_dir=None):
    #     root = get_volttron_root()
    #     services_core = os.path.join(root, "services/core")
    #     if not agent_dir:
    #         return services_core
    #     return os.path.join(services_core, agent_dir)

    # def get_ops(agent_dir=None):
    #     root = get_volttron_root()
    #     ops_dir = os.path.join(root, "services/ops")
    #     if not agent_dir:
    #         return ops_dir
    #     return os.path.join(ops_dir, agent_dir)

    # def get_examples(agent_dir):
    #     root = get_volttron_root()
    #     examples_dir = os.path.join(root, "examples")
    #     if not agent_dir:
    #         return examples_dir
    #     return os.path.join(examples_dir, agent_dir)

    # def is_instance_running(volttron_home=None):

    #     if volttron_home is None:
    #         volttron_home = get_home()

    #     instance_file = os.path.expanduser("~/.volttron_instances")
    #     if not os.path.isfile(instance_file):
    #         return False

    #     with open(instance_file, "r") as fp:
    #         jsonobj = jsonapi.loads(fp.read())

    #     if volttron_home not in jsonobj:
    #         return False

    #     obj = jsonobj[volttron_home]
    #     pid = obj.get("pid", None)

    #     if not pid:
    #         return False

    #     return psutil.pid_exists(pid)

    # def is_rabbitmq_available():
    #     rabbitmq_available = True
    #     try:
    #         import pika

    #         rabbitmq_available = True
    #     except ImportError:
    #         os.environ["RABBITMQ_NOT_AVAILABLE"] = "True"
    #         rabbitmq_available = False
    #     return rabbitmq_available

    # __config__ = None

    # def get_platform_config():
    #     global __config__
    #     if os.environ.get("VOLTTRON_HOME") is None:
    #         raise Exception("VOLTTRON_HOME must be specified before calling this function.")

    #     if __config__ is None:
    #         __config__ = FrozenDict()
    #         volttron_home = get_home()
    #         config_file = os.path.join(volttron_home, "config")
    #         if os.path.exists(config_file):
    #             parser = ConfigParser()
    #             parser.read(config_file)
    #             options = parser.options("volttron")
    #             for option in options:
    #                 __config__[option] = parser.get("volttron", option)
    #             __config__.freeze()
    #     return __config__

    # def update_platform_config(values: dict) -> None:
    #     global __config__

    #     if __config__ is None:
    #         cfg = get_platform_config()
    #     else:
    #         cfg = __config__
    #         # Make sure we can update items
    #         cfg._frozen = False

    #     cfg.update(values)

    #     config_file = get_config_path()
    #     with open(config_file, "w") as fp:
    #         p = ConfigParser()
    #         p.add_section("volttron")
    #         for k, v in cfg.items():
    #             p.set("volttron", k, v)

    #         cfg.freeze()
    #         p.write(fp)

    #     return get_platform_config()

    # def build_vip_address_string(vip_root, serverkey, publickey, secretkey):
    #     """Build a full vip address string based upon the passed arguments

    #     All arguments are required to be non-None in order for the string to be
    #     created successfully.

    #     :raises ValueError if one of the parameters is None.
    #     """
    #     _log.debug(
    #         "root: {}, serverkey: {}, publickey: {}, secretkey: {}".format(
    #             vip_root, serverkey, publickey, secretkey
    #         )
    #     )
    #     parsed = urlparse(vip_root)
    #     if parsed.scheme == "tcp":
    #         if not (serverkey and publickey and secretkey and vip_root):
    #             raise ValueError("All parameters must be entered.")

    #         root = "{}?serverkey={}&publickey={}&secretkey={}".format(
    #             vip_root, serverkey, publickey, secretkey
    #         )

    #     elif parsed.scheme == "ipc":
    #         root = vip_root
    #     else:
    #         raise ValueError("Invalid vip root specified!")

    #     return root

    # def update_volttron_script_path(path: str) -> str:
    #     """
    #     Assumes that path's current working directory is in the root directory of the volttron codebase.

    #     Prepend 'VOLTTRON_ROOT' to internal volttron script if 'VOLTTRON_ROOT' is set and return new path;
    #     otherwise, return original path
    #     :param path: relative path to the internal volttron script
    #     :return: updated path to volttron script
    #     """
    #     if os.environ["VOLTTRON_ROOT"]:
    #         args = path.split("/")
    #         path = f"{os.path.join(os.environ['VOLTTRON_ROOT'], *args)}"
    #     _log.debug(f"Path to script: {path}")
    #     return path



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import argparse
import base64
import glob
import hashlib
import logging
import os
import shutil
from pathlib import Path
import sys
import tempfile
from typing import Callable

import gevent
import yaml

from volttron.client.vip.agent.results import AsyncResult
from volttron.types.agent_context import AgentInstallOptions

from volttron.utils import execute_command
from volttron.utils import jsonapi


class InstallRuntimeError(RuntimeError):
    pass


from volttron.client.logs import get_logger

_log = get_logger()

_stdout = sys.stdout
_stderr = sys.stderr


def _build_from_pyproject(install_path: Path) -> Path:
    """build project from poetry based upon the pyproject.toml file.

    This is a local build when the user passes a directory to the install command.

    :param install_path: Path to the directory containing the pyproject.toml file.
    :type install_path: Path
    :raises InstallRuntimeError: If no wheel file was built during the built process this is thrown.
    :return: Path to the wheel file that was built.
    :rtype: Path
    """

    pyproject_path = install_path / "pyproject.toml"
    assert pyproject_path.exists(), f"pyproject.toml not found in {install_path}"

    dist_path: Path = install_path / "dist"

    cmd = ["poetry", "build", "-vv"]
    output = execute_command(cmd, cwd=install_path)
    match = sorted(dist_path.glob("*.whl"))

    if match:
        return Path(match[-1])
    else:
        raise InstallRuntimeError(
            f"No .whl file found in {dist_path} after running command {' '.join(cmd)}. "
            f"\nCommand returned stdout:\n{output}")


def _install_and_initialize_agent(opts: argparse.Namespace,
                                  wheel_file: Path = None,
                                  pypi_string: str = None):
    """Create a new agent on the platform.

    This is the main installation function for sending/installing an agent on the platform. Depending
    on the arguments, the agent will be installed from a wheel file or from pypi.  The options will be
    retrieved from the opts namespace argument.

    This function creates an `AgentInstallOptions` object and sends it to the platform.  The platform
    will then install the agent and return the agent uuid.

    :param opts: The command line options for the install command.
    :type opts: argparse.Namespace
    :param wheel_file: If a path is specified ".whl" file will be sent, defaults to None
    :type wheel_file: Path, optional
    :param pypi_string: A string to install from pypi, defaults to None
    :type pypi_string: str, optional
    :raises InstallRuntimeError: Raised if a problem with a passed config file.
    :raises ValueError: Raised if the agent was not installed properly.
    """

    assert opts.connection, "Connection must have been created to access this feature."
    assert wheel_file or pypi_string, "Either a wheel file or pypi string must be specified."
    assert not (wheel_file
                and pypi_string), "Only one of wheel_file or pypi_string can be specified."

    connection = opts.connection

    if wheel_file and not wheel_file.exists():
        raise InstallRuntimeError(f"Wheel file {wheel_file} does not exist!")

    # Verify and load agent_config up from the opts.  agent_config will
    # be a yaml config file.
    agent_config = opts.agent_config
    if agent_config is None:
        agent_config = {}

    cfg = None    # temp file if agent_config is a dict
    # if not a dict then config should be a filename
    if not isinstance(agent_config, dict):
        config_file = Path(agent_config).expanduser().as_posix()
        if not Path(config_file).exists():
            raise InstallRuntimeError(f"Config file {config_file} does not exist!")
    else:
        cfg = tempfile.NamedTemporaryFile()
        with open(cfg.name, "w") as fout:
            fout.write(yaml.safe_dump(agent_config))
        config_file = cfg.name

    try:
        with open(config_file) as fp:
            config_dict = yaml.safe_load(fp)
    except Exception as exc:
        raise InstallRuntimeError(exc)
    finally:
        if cfg:
            cfg.close()

    agent = None    # holds the wheel file or the pypi string
    agent_data = None    # Holds the base64 encoded data of the wheel file.
    if wheel_file:
        with open(wheel_file, "rb") as fp:
            agent_data = base64.b64encode(fp.read()).decode("utf-8")
        agent = wheel_file.name
    else:
        agent = pypi_string

    agent_install = AgentInstallOptions(source=agent,
                                        identity=opts.vip_identity,
                                        data=agent_data,
                                        agent_config=config_dict,
                                        force=opts.force,
                                        allow_prerelease=opts.pre_release)

    agent_uuid = connection.call("install_agent", agent_install.to_dict())

    if not agent_uuid:
        raise ValueError(f"Agent was not installed properly.")

    if isinstance(agent_uuid, AsyncResult):
        agent_uuid = agent_uuid.get()

    output_dict = dict(agent_uuid=agent_uuid)

    if opts.tag:
        _log.debug(f"Tagging agent {agent_uuid}, {opts.tag}")
        opts.connection.call("tag_agent", agent_uuid, opts.tag)
        output_dict["tag"] = opts.tag

    if opts.enable or opts.priority != -1:
        output_dict["enabling"] = True
        if opts.priority == -1:
            opts.priority = "50"
        _log.debug(f"Prioritinzing agent {agent_uuid},{opts.priority}")
        output_dict["priority"] = opts.priority

        opts.connection.call("prioritize_agent", agent_uuid, str(opts.priority))

    try:

        if opts.start:
            gevent.sleep(2)
            _log.debug(f"Staring agent {agent_uuid}")
            opts.connection.call("start_agent", agent_uuid)
            output_dict["starting"] = True

            _log.debug(f"Getting agent status {agent_uuid}")
            gevent.sleep(opts.agent_start_time)
            status = opts.connection.call("agent_status", agent_uuid)
            if status[0] is not None and status[1] is None:
                output_dict["started"] = True
                output_dict["pid"] = status[0]
            else:
                output_dict["started"] = False
            _log.debug(f"Status returned {status}")
    except Exception as e:
        _log.error(e)

    if opts.json:
        sys.stdout.write("%s\n" % jsonapi.dumps(output_dict, indent=4))
    else:
        if output_dict.get("started"):
            sys.stdout.write(f"Agent {agent_uuid} installed and started [{output_dict['pid']}]\n")
        else:
            sys.stdout.write(f"Agent {agent_uuid} installed\n")
    if opts.csv:
        keylen = len(output_dict)
        keyline = ""
        valueline = ""
        keys = list(output_dict.keys())
        for k in range(keylen):
            if k < keylen - 1:
                keyline += "%s," % keys[k]
                valueline += "%s," % output_dict[keys[k]]
            else:
                keyline += "%s" % keys[k]
                valueline += "%s" % output_dict[keys[k]]
        sys.stdout.write("%s\n%s\n" % (keyline, valueline))


def install_agent_vctl(opts: argparse.Namespace, callback=None):
    """
    The `install_agent_vctl` function is called from the volttron-ctl or vctl install
    sub-parser.

    This function uses the `opts` namespace install_path and wheel attributes to install
    the agent on the connected platform instance.  The `opts`.connection attribute must
    be set and connected prior to calling this function.

    If an error occurs during this function it will be passed up the stack and the callback
    function will not be called.

    :param opts: The namespace object containing the install_path and wheel attributes.
    :type opts: argparse.Namespace
    :param callback: A callback function to call after the agent is successfullly installed.
    :type callback: function
    """

    assert opts.connection, "Connection must have been created to access this feature."

    try:
        install_path = Path(opts.install_path)
    except AttributeError:
        install_path = Path(opts.wheel)

    if install_path.is_dir():
        print(f"Building agent from {install_path}")
        install_path = _build_from_pyproject(install_path)

    if install_path.is_file() and install_path.suffix == ".whl":
        print(f"Installing from wheel {install_path}")
        _install_and_initialize_agent(opts, wheel_file=install_path)
    else:
        print(f"Installing from pypi {install_path}")
        _install_and_initialize_agent(opts, pypi_string=install_path)

    if callback:
        callback()


def add_install_agent_parser(add_parser_fn: Callable):
    """Creat and add the parser for the install command.

    :param add_parser_fn: A
    :type add_parser_fn: _type_
    """
    install = add_parser_fn(
        "install",
        help="install agent from wheel",
        epilog="Optionally you may specify the --tag argument to tag the "
        "agent during install without requiring a separate call to "
        "the tag command. ",
    )
    install.add_argument(
        "--skip-requirements",
        help=
        "Skip installing requirements from a requirements.txt if present in the agent directory.",
    )
    install.add_argument(
        "install_path",
        help="path to agent wheel or directory for agent installation",
    )
    install.add_argument("--tag", help="tag for the installed agent")
    install.add_argument(
        "--vip-identity",
        help="VIP IDENTITY for the installed agent. "
        "Overrides any previously configured VIP IDENTITY.",
    )
    install.add_argument("--agent-config", help="Agent configuration!")
    install.add_argument(
        "-f",
        "--force",
        action="store_true",
        help=
        "agents are uninstalled by tag so force allows multiple agents to be removed at one go.",
    )
    install.add_argument(
        "--priority",
        default=-1,
        type=int,
        help="priority of startup during instance startup",
    )
    install.add_argument(
        "--start",
        action="store_true",
        help="start the agent during the script execution",
    )
    install.add_argument(
        "--enable",
        action="store_true",
        help="enable the agent with default 50 priority unless --priority set",
    )
    install.add_argument(
        "--csv",
        action="store_true",
        help="format the standard out output to csv",
    )
    install.add_argument(
        "--json",
        action="store_true",
        help="format the standard out output to json",
    )
    install.add_argument(
        "-st",
        "--agent-start-time",
        default=5,
        type=int,
        help="the amount of time to wait and verify that the agent has started up.",
    )
    install.add_argument(
        "--pre-release",
        "--pre",
        "--allow-prereleases",
        action="store_true",
        help="enables installation of pre-releases and development releases",
    )

    install.set_defaults(func=install_agent_vctl, verify_agents=True)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import argparse
import logging
import os

# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import inspect
import logging
import os
import stat
import sys
import syslog
import traceback
import warnings

from volttron.utils import jsonapi
from volttron.client.logs import AgentFormatter

try:
    HAS_SYSLOG = True
    import syslog
except ImportError:
    HAS_SYSLOG = False

# Keep the ability to have system log output for linux
# this will fail on windows because no syslog.
if HAS_SYSLOG:

    class SyslogFormatter(logging.Formatter):
        _level_map = {
            logging.DEBUG: syslog.LOG_DEBUG,
            logging.INFO: syslog.LOG_INFO,
            logging.WARNING: syslog.LOG_WARNING,
            logging.ERROR: syslog.LOG_ERR,
            logging.CRITICAL: syslog.LOG_CRIT,
        }

        def format(self, record):
            level = self._level_map.get(record.levelno, syslog.LOG_INFO)
            return "<{}>".format(level) + super(SyslogFormatter, self).format(record)


def get_default_loggers_config(level: int) -> dict:
    level_server = logging.ERROR
    level_client = logging.WARNING
    match level:
        case 1:
            level_server = logging.WARNING
            level_client = logging.INFO
        case 2:
            level_server = logging.INFO
            level_client = logging.DEBUG
        case 3:
            level_client = logging.DEBUG
            level_server = logging.DEBUG

    return {
        "volttron.messagebus": {
            "level": logging.getLevelName(level_server)
        },
        "volttron.server": {
            "level": logging.getLevelName(level_server)
        },
        "volttron.services": {
            "level": logging.getLevelName(level_server)
        },
        "volttron.client": {
            "level": logging.getLevelName(level_client)
        }
    }


def get_default_logging_config(level: int = logging.WARNING) -> dict:
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {
                "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": level,
                "formatter": "simple",
                "stream": "ext://sys.stdout"
            }
    # ,
    # "file": {
    #     "class": "logging.FileHandler",
    #     "level": "INFO",
    #     "formatter": "simple",
    #     "filename": "myapp.log",
    #     "mode": "a"
    # }
        },
        "loggers": get_default_loggers_config(level),
        "root": {
            "level": level,
            "handlers": ["console"]
        }
    }


def isapipe(fd):
    fd = getattr(fd, "fileno", lambda: fd)()
    return stat.S_ISFIFO(os.fstat(fd).st_mode)


__enable_trace__ = False


def enable_trace():
    global __enable_trace__
    __enable_trace__ = True


def disable_trace():
    global __enable_trace__
    __enable_trace__ = False


def logtrace(func: callable, *args, **kwargs):
    """
    Decorator that logs the function call and return value.

    Example:
        @logtrace
        def add(a, b):
            return a + b

        add(2, 3)
        # Output in debug log:
        # add(a, b) called with (2, 3), {}
        # add returned: 5

    @param func: The function to be decorated.
    @type func: callable
    @return: The decorated function.
    @rtype: callable
    """
    enabled = kwargs.pop('enabled', False)
    logger = logging.getLogger(func.__module__)
    sig = inspect.signature(func)

    def do_logging(*args, **kwargs):
        if __enable_trace__:
            logger.debug(f"-->{func.__name__}{sig} called with {args}, {kwargs}")
        ret = func(*args, **kwargs)
        if __enable_trace__:
            logger.debug(f"<--{func.__name__} returned: {ret}")
        return ret

    return do_logging


class JsonFormatter(logging.Formatter):

    def format(self, record):
        dct = record.__dict__.copy()
        dct["msg"] = record.getMessage()
        dct.pop("args")
        exc_info = dct.pop("exc_info", None)
        if exc_info:
            dct["exc_text"] = "".join(traceback.format_exception(*exc_info))
        return jsonapi.dumps(dct)


class FramesFormatter(object):

    def __init__(self, frames):
        self.frames = frames

    def __repr__(self):
        output = ''
        for f in self.frames:
            output += str(f)
        return output

    __str__ = __repr__


def log_to_file(file, level=logging.WARNING, handler_class=logging.StreamHandler):
    """Direct log output to a file (or something like one)."""
    handler = handler_class(file)
    handler.setLevel(level)
    handler.setFormatter(AgentFormatter("%(asctime)s %(composite_name)s %(levelname)s: %(message)s"))
    root = logging.getLogger()
    root.setLevel(level)
    root.addHandler(handler)


# def setup_logging(level=logging.DEBUG, console=False):
#     root = logging.getLogger()
#     if not root.handlers:
#         handler = logging.StreamHandler()

#         if isapipe(sys.stderr) and "_LAUNCHED_BY_PLATFORM" in os.environ:
#             handler.setFormatter(JsonFormatter())
#         elif console:
#             # Below format is more readable for console
#             handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
#         else:
#             fmt = "%(asctime)s %(name)s %(levelname)s: %(message)s"
#             handler.setFormatter(logging.Formatter(fmt))
#         if level != logging.DEBUG:
#             # import it here so that when urllib3 imports the requests package, ssl would already got
#             # monkey patched by gevent.
#             # and this warning is needed only when log level is not debug
#             from urllib3.exceptions import InsecureRequestWarning

#             warnings.filterwarnings("ignore", category=InsecureRequestWarning)
#         root.addHandler(handler)
#     root.setLevel(level)


def log_to_file(file_, level=logging.WARNING, handler_class=logging.StreamHandler):
    """
    Direct log output to a file (or something like one).
    """
    handler = handler_class(file_)
    handler.setLevel(level)
    if "VOLTTRON_SERVER" in os.environ:
        format_str = "%(asctime)s %(composite_name)s(%(lineno)d) SERVER %(levelname)s: %(message)s"
    else:
        format_str = "%(asctime)s %(composite_name)s(%(lineno)d) %(levelname)s: %(message)s"
    handler.setFormatter(AgentFormatter(fmt=format_str))
    root = logging.getLogger()
    if root.level < level:
        root.setLevel(level)
    root.handlers.clear()
    root.addHandler(handler)


def configure_logging(conf_path):
    """
    Load logging configuration from a file.

    Several formats are possible: ini, JSON, Python, and YAML. The ini
    format uses the standard Windows ini file format and is read in
    using logging.config.fileConfig(). The remaining formats will be
    read in according to the serialization format and the resulting
    dictionary will be passed to logging.config.dictConfig(). See the
    logging.config module for specifics on the two file and dict
    formats. Returns None on success, (path, exception) on error.

    The default format is ini. Other formats will be selected based on
    the file extension. Each format can be forced, regardless of file
    extension, by prepending the path with the format name followed by a
    colon:

      Examples:
        config.json        is loaded as JSON
        config.conf        is loaded as ini
        json:config.conf   is loaded as JSON

    YAML formatted configuration files require the PyYAML package.
    """

    conf_format = "ini"
    if conf_path.startswith("ini:"):
        conf_format, conf_path = "ini", conf_path[4:]
    elif conf_path.startswith("json:"):
        conf_format, conf_path = "json", conf_path[5:]
    elif conf_path.startswith("py:"):
        conf_format, conf_path = "py", conf_path[3:]
    elif conf_path.startswith("yaml:"):
        conf_format, conf_path = "yaml", conf_path[5:]
    elif conf_path.endswith(".json"):
        conf_format = "json"
    elif conf_path.endswith(".py"):
        conf_format = "py"
    elif conf_path.endswith(".yaml"):
        conf_format = "yaml"

    if conf_format == "ini":
        try:
            logging.config.fileConfig(conf_path)
        except (ValueError, TypeError, AttributeError, ImportError) as exc:
            return conf_path, exc
        return

    with open(conf_path) as conf_file:
        if conf_format == "json":
            try:
                conf_dict = jsonapi.load(conf_file)
            except ValueError as exc:
                return conf_path, exc
        elif conf_format == "py":
            import ast

            try:
                conf_dict = ast.literal_eval(conf_file.read())
            except ValueError as exc:
                return conf_path, exc
        else:
            try:
                import yaml
            except ImportError:
                return (
                    conf_path,
                    "PyYAML must be installed before "
                    "loading logging configuration from a YAML file.",
                )
            try:
                conf_dict = yaml.safe_load(conf_file.read())
            except yaml.YAMLError as exc:
                return conf_path, exc
    try:
        logging.config.dictConfig(conf_dict)
    except (ValueError, TypeError, AttributeError, ImportError) as exc:
        return conf_path, exc


class LogLevelAction(argparse.Action):
    """
    Action to set the log level of individual modules.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        for pair in values.split(","):
            if not pair.strip():
                continue
            try:
                logger_name, level_name = pair.rsplit(":", 1)
            except (ValueError, TypeError):
                raise argparse.ArgumentError(self, "invalid log level pair: {}".format(values))
            try:
                level = int(level_name)
            except (ValueError, TypeError):
                try:
                    level = getattr(logging, level_name)
                except AttributeError:
                    raise argparse.ArgumentError(self, "invalid log level {!r}".format(level_name))
            logger = logging.getLogger(logger_name)
            logger.setLevel(level)



from __future__ import annotations

import importlib
import inspect
import logging
import sys
import traceback
from pathlib import Path
from types import ModuleType
from typing import List, Type

from volttron.utils.logs import logtrace

__all__: List[str] = ["get_module", "get_class", "get_subclasses", "load_dir"]

_log = logging.getLogger(__name__)

__loaded__: set[str] = set()


@logtrace
def load_dir(package: str, pth: Path):
    """
    Recursively loads all modules within a directory.

    :param package: The package name.
    :type package: str
    :param pth: The path to the directory.
    :type pth: :class:`pathlib.Path`

    This function recursively loads all modules within the specified directory. It iterates
    over the files and subdirectories in the directory, imports each module, and loads any
    subdirectories as sub-packages.

    If a file is a Python source file (`.py`), it is imported as a module. If a file is a
    directory, it is loaded as a sub-package.

    Example usage::

        load_dir('volttron', Path('/path/to/directory'))

    .. note::
        This function does not load compiled Python files (`.pyc`) or files in
        the `__pycache__` directory.

    .. warning::
        Be cautious when using this function, as it imports and loads all modules within a
        directory, which may have unintended side effects.
    """

    # Find the caller of this function
    caller_module = inspect.currentframe().f_back.f_globals["__name__"]

    get_mod_name = lambda pth: pth.name if pth.is_dir() else pth.stem

    # We only want to load from py files rather than pyc files
    # so filter
    for p in filter(lambda x: x.name != '__pycache__', pth.iterdir()):

        new_package = f"{package}.{get_mod_name(p)}"

        if new_package.endswith("__init__") or new_package.endswith('.mypy_cache'):
            _log.debug(f"Skipping {new_package}")
            continue

        if f"{caller_module}.__init__" == new_package:
            _log.debug(f"Skipping {new_package}")
            continue
        # print(globals())
        if p.absolute().as_posix() not in __loaded__:
            __loaded__.add(p.absolute().as_posix())
            _log.debug(f"Loading {new_package}")
            try:
                importlib.import_module(new_package)
            except ImportError as ex:
                traceback.print_exc()
                _log.error(f"Failed to import {new_package} due to {ex}")
                raise
                continue
            # after = set(globals().keys())
            # print("After import")
            # print(before - after)
            # print("AFTER!!!!")
            # for k in list(globals().keys()):
            #     print(k)

            # assert new_package in globals()
            if p.is_dir():
                load_dir(new_package, p)



from typing import Any

from gevent import Greenlet

from volttron.client.vip.agent import Agent
from volttron.types import Service
from volttron.types.message import Message
from volttron.types.server_config import ServerConfig


class ServiceInterface(Agent):

    def __init__(self, server_config: ServerConfig = None, **kwargs):
        self._server_config = server_config
        super().__init__(**kwargs)

    @classmethod
    def get_kwargs_defaults(cls) -> dict[str, Any]:
        """
        Class method that allows the specific class to have the ability to specify
        what service arguments are available as defaults.
        """
        return {}

    def spawn_in_greenlet(self) -> Greenlet:
        """
        Start the execution of a volttron agent.
        """
        event = gevent.event.Event()
        task = gevent.spawn(self.core.run, event)
        event.wait()
        del event
        return task



import functools
from dataclasses import dataclass, field
from dataclass_wizard import JSONSerializable

from volttron.types.auth.auth_credentials import Credentials
from volttron.types import Identity, Tag


@dataclass
class AgentOptions:
    heartbeat_autostart: bool = False
    '''If True, the heartbeat will start automatically'''

    heartbeat_period: int = 60
    '''If the heartbeat is started, this is the period in seconds between heartbeats'''

    volttron_home: str = None
    '''A setting for the volttron home directory, set by the platform if not set'''

    agent_uuid: str = None
    '''An installed agent's uuid'''

    enable_store: bool = True
    '''If True, the agent will have a config store and be able to store configurations'''

    reconnect_interval: int = None

    version: str = "0.1"
    volttron_central_address: str = None,
    volttron_central_instance_name: str = None,
    tag_vip_id: str = None,
    tag_refresh_interval: int = -1
    custom_options: dict[str, any] = field(default_factory=dict)


@dataclass(frozen=True)
class AgentContext:
    credentials: Credentials
    address: str | list[str]
    options: AgentOptions = field(default_factory=lambda: AgentOptions())

    # Use functools rather than try to modify the property of
    # frozen class
    @functools.cached_property
    def address(self):
        if isinstance(self.address, str):
            return [self.address]
        else:
            return self.address


@dataclass
class AgentInstallOptions(JSONSerializable):
    """ A metaclass for controling the installation of an agent onto the platform.

    Parameters
    ----------
    source : str
        The name of the agent to install from pypi or the name of the
        wheel file.  If the wheel file, then data parameter must be included.  If a wheel
        file the source MUST end with whl.
    start : bool = False
        Should the agent be started after installing successfully.
    data : str | None
        If source is specified as a .whl file, the data associated with the wheel file
        MUST be set in this variable.
    identity : Identity | None
        The identity of the installed agent.  If identity is not specified, then the
        name of the source will be used to create the agent.
    tag : Tag | None
        A tag associated with the agent
    force : bool = False
        If the identity is already on the platform, should we overwrite it or not
    allow_prerelease: bool = False
        Should the platform allow pre-release artifacts or only releases from pypi.
    agent_config : dict = {}
        A default configuration for the agent to be installed.  This will be stored
        so that it is passed to the agent on startup.
    credentials : Credentials | None
        A set of credentials that should be used for this agent.  In general, the platform
        will generate credentials as the agent is installed, but this method allows the
        calling platform to specify them externally.  Make sure the passed credentials are
        available on the platform from the CredentialsFactory or this method will fail.
    enabled : bool = False
        An enabled agent will start when the platform is started.  If enabled and priority
        is not set then a 50 priority will be set.
    priority: int | None
        A priority specified means the agent is enabled.  The number should be between
        1-99 where lower numbers start first.

    Raises
    ------
    ValueError if there is an issue with any of the arguments passed to the class.

    Returns
    -------
    An agent uuid if the agent is successful
    """
    source: str
    start: bool = False
    data: str | None = None
    identity: Identity | None = None
    tag: Tag | None = None
    force: bool = False
    allow_prerelease: bool = False
    agent_config: dict = field(default_factory=dict)
    credentials: Credentials | None = None
    enabled: bool = False
    priority: int | None = None

    def __post_init__(self):

        if self.priority is not None:
            if self.priority < 0 or self.priority > 99:
                raise ValueError(f"Invalid priority specified!")
            self.enabled = True

        if self.enabled and self.priority is None:
            self.priority = 50



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

__all__ = [
    "format_timestamp",
    "parse_timestamp_string",
    "get_aware_utc_now",
    "get_utc_seconds_from_epoch",
    "process_timestamp",
    "fix_sqlite3_datetime",
]

import calendar
from datetime import datetime
from dateutil.parser import parse
from dateutil.tz import tzutc, tzoffset
import logging
import pytz
from tzlocal import get_localzone

_log = logging.getLogger(__name__)


def format_timestamp(time_stamp):
    """Create a consistent datetime string representation based on
    ISO 8601 format.

    YYYY-MM-DDTHH:MM:SS.mmmmmm for unaware datetime objects.
    YYYY-MM-DDTHH:MM:SS.mmmmmm+HH:MM for aware datetime objects

    :param time_stamp: value to convert
    :type time_stamp: datetime
    :returns: datetime in string format
    :rtype: str
    """

    time_str = time_stamp.strftime("%Y-%m-%dT%H:%M:%S.%f")

    if time_stamp.tzinfo is not None:
        sign = "+"
        td = time_stamp.tzinfo.utcoffset(time_stamp)
        if td.days < 0:
            sign = "-"
            td = -td

        seconds = td.seconds
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        time_str += "{sign}{HH:02}:{MM:02}".format(sign=sign, HH=hours, MM=minutes)

    return time_str


def parse_timestamp_string(time_stamp_str):
    """
    Create a datetime object from the supplied date/time string.
    Uses dateutil.parse with no extra parameters.

    For performance reasons we try
    YYYY-MM-DDTHH:MM:SS.mmmmmm
    or
    YYYY-MM-DDTHH:MM:SS.mmmmmm+HH:MM
    based on the string length before falling back to dateutil.parse.

    @param time_stamp_str:
    @return: value to convert
    """

    if len(time_stamp_str) == 26:
        try:
            return datetime.strptime(time_stamp_str, "%Y-%m-%dT%H:%M:%S.%f")
        except ValueError:
            pass

    elif len(time_stamp_str) == 32:
        try:
            base_time_stamp_str = time_stamp_str[:26]
            time_zone_str = time_stamp_str[26:]
            time_stamp = datetime.strptime(base_time_stamp_str, "%Y-%m-%dT%H:%M:%S.%f")
            # Handle most common case.
            if time_zone_str == "+00:00":
                return time_stamp.replace(tzinfo=pytz.UTC)

            hours_offset = int(time_zone_str[1:3])
            minutes_offset = int(time_zone_str[4:6])

            seconds_offset = hours_offset * 3600 + minutes_offset * 60
            if time_zone_str[0] == "-":
                seconds_offset = -seconds_offset

            return time_stamp.replace(tzinfo=tzoffset("", seconds_offset))

        except ValueError:
            pass

    return parse(time_stamp_str)


def get_aware_utc_now():
    """Create a timezone aware UTC datetime object from the system time.

    :returns: an aware UTC datetime object
    :rtype: datetime
    """
    utcnow = datetime.utcnow()
    utcnow = pytz.UTC.localize(utcnow)
    return utcnow


def get_utc_seconds_from_epoch(timestamp=None):
    """
    convert a given time stamp to seconds from epoch based on utc time. If
    given time is naive datetime it is considered be local to where this
    code is running.
    @param timestamp: datetime object
    @return: seconds from epoch
    """

    if timestamp is None:
        timestamp = datetime.now(tz=tzutc())

    if timestamp.tzinfo is None:
        local_tz = get_localzone()

        # Note:
        # We replace the time zone here which allows us to get the timezone and set it to the
        # current local timezone.  This may have an issue when we are in the daylight savings time
        # era.  See using fold on the timestamp for fixing this.
        #
        # https://pytz-deprecation-shim.readthedocs.io/en/latest/migration.html#acquiring-a-tzinfo-object
        #
        # TODO: Handle timestamp without using localize for migration.
        timestamp = timestamp.replace(tzinfo=local_tz)

    # utctimetuple can be called on aware timestamps and it will
    # convert to UTC first.
    seconds_from_epoch = calendar.timegm(timestamp.utctimetuple())
    # timetuple loses microsecond accuracy so we have to put it back.
    seconds_from_epoch += timestamp.microsecond / 1000000.0
    return seconds_from_epoch


def process_timestamp(timestamp_string, topic=""):
    """
    Convert timestamp string timezone aware utc timestamp
    @param timestamp_string: datetime string to parse
    @param topic: topic to which parse errors are published
    @return: UTC datetime object and the original timezone of input datetime
    """
    if timestamp_string is None:
        _log.error("message for {topic} missing timetamp".format(topic=topic))
        return

    try:
        timestamp = parse_timestamp_string(timestamp_string)
    except (ValueError, TypeError):
        _log.error("message for {topic} bad timetamp string: {ts_string}".format(
            topic=topic, ts_string=timestamp_string))
        return

    if timestamp.tzinfo is None:
        timestamp = timestamp.replace(tzinfo=pytz.UTC)
        original_tz = None
    else:
        original_tz = timestamp.tzinfo
        timestamp = timestamp.astimezone(pytz.UTC)
    return timestamp, original_tz


def fix_sqlite3_datetime(sql=None):
    """Primarily for fixing the base historian cache on certain versions
    of python.

    Registers a new datetime converter to that uses dateutil parse. This
    should
    better resolve #216, #174, and #91 without the goofy workarounds that
    change data.

    Optional sql argument is for testing only.
    """
    if sql is None:
        import sqlite3 as sql

    def parse(time_stamp_bytes):
        return parse_timestamp_string(time_stamp_bytes.decode("utf-8"))

    sql.register_adapter(datetime, format_timestamp)
    sql.register_converter("timestamp", parse)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import weakref

from volttron.client.messaging import topics
from volttron.client.messaging.headers import DATE
from volttron.client.messaging.health import *
from volttron.client.vip.agent.subsystems.base import SubsystemBase
from volttron.utils.context import ClientContext as cc
"""
The health subsystem allows an agent to store it's health in a non-intrusive
way.
"""

__docformat__ = "reStructuredText"
__version__ = "1.1"

from volttron.client.logs import get_logger

_log = get_logger()


class Health(SubsystemBase):

    def __init__(self, owner, core, rpc):
        self._owner = owner
        self._core = weakref.ref(core)
        self._rpc = weakref.ref(rpc)
        self._statusobj = Status.build(STATUS_GOOD, status_changed_callback=self._status_changed)
        self._status_callbacks = set()

        def onsetup(sender, **kwargs):
            rpc.export(self.set_status, "health.set_status")
            rpc.export(self.get_status, "health.get_status")
            rpc.export(self.get_status, "health.get_status_json")
            rpc.export(self.send_alert, "health.send_alert")

        core.onsetup.connect(onsetup, self)

    def send_alert(self, alert_key: str, statusobj: Status):
        """
        An alert_key is a quasi-unique key.  A listener to the alert can
        determine whether to pass the alert on to a higher level based upon
        the frequency of this alert.

        :param statusobj: The status for the alert.
        :param alert_key:
        :return:
        """
        if not isinstance(statusobj, Status):
            raise ValueError("statusobj must be a Status object.")
        agent_class = self._owner.__class__.__name__
        fq_identity = cc.get_fq_identity(self._core().identity)
        # RMQ and other message buses can't handle '.' because it's used as the separator.  This
        # causes us to change the alert topic's agent_identity to have '_' rather than '.'.
        topic = topics.ALERTS(agent_class=agent_class,
                              agent_identity=fq_identity.replace(".", "_"))
        headers = dict(alert_key=alert_key)

        self._owner.vip.pubsub.publish("pubsub",
                                       topic=topic.format(),
                                       headers=headers,
                                       message=statusobj.as_json()).get(timeout=10)

    def add_status_callback(self, fn):
        """
        Add callbacks to the passed function.  The function must have the
        following interface

        .. code::python

            def status_callback(status, context):

        :param fn: The method to be executed when status is changed.
        :param fn: callable
        """
        self._status_callbacks.add(fn)

    def _status_changed(self):
        """Internal function that happens when the status changes state.
        :return:
        """
        remove = set()
        for fn in self._status_callbacks:
            try:
                fn(self._statusobj.status, self._statusobj.context)
            except NameError:
                remove.add(fn)
        # Removes the items that are in remove from the callbacks.
        self._status_callbacks.difference_update(remove)

        self._owner.vip.heartbeat.restart()

    def set_status(self, status, context=None):
        """RPC method

        Updates the agents status to the new value with the specified context.

        :param: status: str: GODD, BAD
        :param: context: str: A serializable that denotes the context of
        status.
        """
        self._statusobj.update_status(status, context)

    def get_status(self):
        """ "RPC method

        Returns the last updated status from the object with the context.

        The minimum output from the status would be:

            {
                "status": "GOOD",
                "context": None,
                "utc_last_update": "2016-03-31T15:40:32.685138+0000"
            }

        """
        return self._statusobj.as_dict()    # .as_json()

    # TODO fetch status value from status object
    def get_status_value(self):
        return self._statusobj.status

    def get_status_json(self):
        """ "RPC method

        Returns the last updated status from the object with the context.

        The minimum output from the status would be:

            {
                "status": "GOOD",
                "context": None,
                "utc_last_update": "2016-03-31T15:40:32.685138+0000"
            }

        """
        return self._statusobj.as_json()

    # TODO define publish for health messaging
    # TODO fix topic
    # TODO fix self.core
    def publish(self):
        topic = "heartbeat/" + self._core().identity
        headers = {DATE: format_timestamp(get_aware_utc_now())}
        message = self.get_status_value()

        self._owner.vip.pubsub.publish("pubsub", topic, headers, message)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import inspect
import logging
import os
import re
import sys
import traceback
import weakref

import gevent.local
from gevent.event import AsyncResult
from zmq import ZMQError
from zmq.green import ENOTSOCK

from volttron.utils import jsonapi, jsonrpc
from volttron.client.known_identities import AUTH, CONTROL, CONTROL_CONNECTION
from ..decorators import annotate, annotations, dualmethod, spawn
from ..results import ResultsDictionary, counter
from .base import SubsystemBase

__all__ = ["RPC"]

_ROOT_PACKAGE_PATH = (os.path.dirname(__import__(__name__.split(".", 1)[0]).__path__[-1]) + os.sep)
from volttron.client.logs import get_logger

_log = get_logger()


def _isregex(obj):
    return (obj is not None and isinstance(obj, str) and len(obj) > 1 and obj[0] == obj[-1] == "/")


class Dispatcher(jsonrpc.Dispatcher):

    def __init__(self, methods, local):
        super(Dispatcher, self).__init__()
        self.methods = methods
        self.local = local
        self._results = ResultsDictionary()

    def serialize(self, json_obj):
        return jsonapi.dumps(json_obj)

    def deserialize(self, json_string):
        return jsonapi.loads(json_string)

    def batch_call(self, requests):
        methods = []
        results = []
        for notify, method, args, kwargs in requests:
            if notify:
                ident = None
            else:
                result = next(self._results)
                ident = result.ident
                results.append(result)
            methods.append((ident, method, args, kwargs))
        return super(Dispatcher, self).batch_call(methods), results

    def call(self, method, args=None, kwargs=None):
        # pylint: disable=arguments-differ
        result = next(self._results)
        return super(Dispatcher, self).call(result.ident, method, args, kwargs), result

    def result(self, response, ident, value, context=None):
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set(value)

    def error(self, response, ident, code, message, data=None, context=None):
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set_exception(jsonrpc.exception_from_json(code, message, data))

    def exception(self, response, ident, message, context=None):
        # XXX: Should probably wrap exception in RPC specific error
        #      rather than re-raising.
        (
            _,    # exc_type
            exc,
            _,    # exc_tb
        ) = sys.exc_info()    # pylint: disable=unused-variable
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set_exception(exc)

    def method(self, request, ident, name, args, kwargs, batch=None, context=None):
        if kwargs:
            try:
                args, kwargs = kwargs["*args"], kwargs["**kwargs"]
            except KeyError:
                pass
        try:
            method = self.methods[name]
        except KeyError:
            if name == "inspect":
                return {"methods": list(self.methods)}
            elif name.endswith(".inspect"):
                try:
                    method = self.methods[name[:-8]]
                except KeyError:
                    pass
                else:
                    return self._inspect(method)
            raise NotImplementedError(name)
        local = self.local
        local.vip_message = context
        local.request = request
        local.batch = batch
        try:
            return method(*args, **kwargs)
        except Exception as exc:    # pylint: disable=broad-except
            exc_tb = traceback.format_exc()
            _log.error("unhandled exception in JSON-RPC method %r: \n%s", name, exc_tb)
            if getattr(method, "traceback", True):
                exc.exc_info = {"exc_tb": exc_tb}
            raise
        finally:
            del local.vip_message
            del local.request
            del local.batch

    @staticmethod
    def _inspect(method):
        response = {"params": {}}
        signature = inspect.signature(method)
        for p in signature.parameters.values():
            response["params"][p.name] = {"kind": p.kind.name}
            if p.default is not inspect.Parameter.empty:
                response["params"][p.name]["default"] = p.default
            if p.annotation is not inspect.Parameter.empty:
                annotation = (p.annotation.__name__ if type(p.annotation) is type else str(p.annotation))
                response["params"][p.name]["annotation"] = annotation
        doc = inspect.getdoc(method)
        if doc:
            response["doc"] = doc
        try:
            source = inspect.getsourcefile(method)
            cut = len(os.path.commonprefix([_ROOT_PACKAGE_PATH, source]))
            source = source[cut:]
            lineno = inspect.getsourcelines(method)[1]
        except Exception:
            pass
        else:
            response["source"] = {"file": source, "line_number": lineno}
        ret = signature.return_annotation
        if ret is not inspect.Signature.empty:
            response["return"] = \
                ret.__name__ if type(ret) is type else str(ret)
        return response


class RPC(SubsystemBase):

    def __init__(self, core, owner, peerlist_subsys):
        self.core = weakref.ref(core)
        self._owner = owner
        self.context = None
        self._exports = {}
        self._dispatcher = None
        self._counter = counter()
        self._outstanding = weakref.WeakValueDictionary()
        core.register("RPC", self._handle_subsystem, self._handle_error)
        core.register(
            "external_rpc",
            self._handle_external_rpc_subsystem,
            self._handle_error,
        )
        self._isconnected = True
        self.peerlist_subsystem = peerlist_subsys
        self.peer_list = {}
        self._protected_rpcs = None

        def export(member):    # pylint: disable=redefined-outer-name
            for name in annotations(member, set, "__rpc__.exports"):
                self._exports[name] = member

        inspect.getmembers(owner, export)

        def setup(sender, **kwargs):
            # pylint: disable=unused-argument
            self.context = gevent.local.local()
            self._dispatcher = Dispatcher(self._exports, self.context)
            self.export(self._add_protected_rpcs, "rpc.add_protected_rpcs")

        core.onsetup.connect(setup, self)
        core.ondisconnected.connect(self._disconnected)
        core.onconnected.connect(self._connected)

    def _connected(self, sender, **kwargs):
        self._isconnected = True
        # Registering to 'onadd' and 'ondrop' signals to get notified
        # whenever new peer is added/removed
        self.peerlist_subsystem.onadd.connect(self._add_new_peer)
        self.peerlist_subsystem.ondrop.connect(self._drop_new_peer)

    def _disconnected(self, sender, **kwargs):
        self._isconnected = False

    def _add_new_peer(self, sender, **kwargs):
        try:
            peer = kwargs.pop("peer")
            message_bus = kwargs.pop("message_bus")
            self.peer_list[peer] = message_bus
        except KeyError:
            pass

    def _drop_new_peer(self, sender, **kwargs):
        try:
            peer = kwargs.pop("peer")
            self.peer_list.pop(peer)
        except KeyError:
            pass

    def _add_auth_check(self, method):
        """
        Adds an authorization check to verify the calling agent has the
        required capabilities.
        """

        def checked_method(*args, **kwargs):
            calling_user = str(self.context.vip_message.user)
            #method_name = method.__name__
            # method.__name__ will give actual method name, but we want the alias used to export this method
            # ex: export("actual_method_name", "alias_exported_which_will_be_used_by_caller")
            method_name = self.context.vip_message.args[0]['method']
            #args_dict = inspect.getcallargs(method, *args, **kwargs)
            signature = inspect.signature(method)
            bound_args = signature.bind(*args, **kwargs)
            bound_args.apply_defaults()
            args_dict = bound_args.arguments
            # Remove self from args_dict if it exists to avoid sending the entire object across.
            # Fixes Issue https://github.com/eclipse-volttron/volttron-core/issues/198
            remove_self = args_dict.pop("self", None)
            from volttron.types.auth import AuthException
            try:
                self.call(AUTH,
                          method="check_rpc_authorization",
                          identity=calling_user,
                          method_name=f"{self.core().identity}.{method_name}",
                          method_args=args_dict).get(timeout=10)
            except AuthException as e:
                # msg = ("method '{}' requires capabilities {}, but capability {} "
                #        "was provided for user {}").format(method.__name__, required_caps,
                #                                           user_capabilites, user)
                raise jsonrpc.exception_from_json(jsonrpc.UNAUTHORIZED, e.args)
            return method(*args, **kwargs)

        return checked_method

    def _wrap_protected_rpcs(self, protected_rpcs):
        """
        Iterates over exported methods and adds authorization checks
        for protected methods
        """
        if protected_rpcs:
            for method_name in self._exports:
                if method_name in protected_rpcs:
                    self._exports[method_name] = self._add_auth_check(self._exports[method_name])

    def _add_protected_rpcs(self, updated_list: list[str]):
        if not self._protected_rpcs:
            # nothing was there before so set variable and update all
            self._protected_rpcs = updated_list
            self._wrap_protected_rpcs(self._protected_rpcs)
        else:
            newly_protected = set(updated_list) - set(self._protected_rpcs)
            self._wrap_protected_rpcs(newly_protected)

    def _remove_protected_rpcs(self, remove_list: list[str]):
        if self._protected_rpcs is None:
            self._protected_rpcs = self.get_protected_rpcs()
        if self._protected_rpcs:
            for r in remove_list:
                if r in self._protected_rpcs:
                    method = getattr(self._owner, r, None)
                    # Verify the retrieved method object
                    if method and inspect.ismethod(method):
                        self._exports[r] = method
                    else:
                        raise ValueError(f"Method '{r}' not found in the instance or is not a method.")

    @spawn
    def _handle_external_rpc_subsystem(self, message):
        ret_msg = dict()
        operation = message.args[0]
        rpc_msg = message.args[1]    # jsonapi.loads(message.args[1])
        try:
            method_args = rpc_msg["args"]
            # message.args = [method_args]
            message.args = method_args
            for idx, msg in enumerate(message.args):
                if isinstance(msg, str):
                    message.args[idx] = jsonapi.loads(msg)
            dispatch = self._dispatcher.dispatch
            # _log.debug("External RPC IN message args {}".format(message))

            responses = [response for response in (dispatch(msg, message) for msg in message.args) if response]
            # _log.debug("External RPC Responses {}".format(responses))
            if responses:
                message.user = ""
                try:
                    message.peer = ""
                    message.subsystem = "external_rpc"
                    frames = []
                    operation = "send_platform"
                    frames.append(operation)
                    msg = jsonapi.dumps(
                        dict(
                            to_platform=rpc_msg["from_platform"],
                            to_peer=rpc_msg["from_peer"],
                            from_platform=rpc_msg["to_platform"],
                            from_peer=rpc_msg["to_peer"],
                            args=responses,
                        ))
                    frames.append(msg)
                except KeyError:
                    _log.error("External RPC message did not contain "
                               "proper message format")
                message.args = jsonapi.dumps(ret_msg)
                try:
                    self.core().connection.send_vip(
                        peer="",
                        subsystem="external_rpc",
                        args=frames,
                        msg_id=message.id,
                        user=message.user,
                        copy=False,
                    )
                except ZMQError as ex:
                    _log.error("ZMQ error: {}".format(ex))
                    pass
        except KeyError:
            pass

    @spawn
    def _handle_subsystem(self, message):
        dispatch = self._dispatcher.dispatch

        responses = [response for response in (dispatch(msg, message) for msg in message.args) if response]
        if responses:
            message.user = ""
            message.args = responses
            try:
                if self._isconnected:
                    self.core().connection.send_vip_object(message, copy=False)

            except ZMQError as exc:
                if exc.errno == ENOTSOCK:
                    _log.debug("Socket send on non-socket %s", self.core().identity)

    def _handle_error(self, sender, message, error, **kwargs):
        result = self._outstanding.pop(message.id, None)
        if isinstance(result, AsyncResult):
            result.set_exception(error)
        elif result:
            for result in result:
                result.set_exception(error)

    def get_exports(self):
        """Returns a list copy of all exported methods."""
        return [method for method in self._exports].copy()

    @dualmethod
    def export(self, method, name=None):
        self._exports[name or method.__name__] = method
        return method

    @export.classmethod
    def export(cls, name=None):    # pylint: disable=no-self-argument
        if name is not None and not isinstance(name, str):
            method, name = name, name.__name__
            annotate(method, set, "__rpc__.exports", name)
            return method

        def decorate(method):
            annotate(method, set, "__rpc__.exports", name)
            return method

        return decorate

    def batch(self, peer, requests):
        request, results = self._dispatcher.batch_call(requests)
        if results:
            items = weakref.WeakSet(results)
            ident = "%s.%s" % (next(self._counter), id(items))
            for result in results:
                result._weak_set = items    # pylint: disable=protected-access
            self._outstanding[ident] = items
        else:
            ident = ""
        if request:
            if self._isconnected:
                try:
                    self.core().connection.send_vip(peer, "RPC", [request], msg_id=ident)
                except ZMQError as exc:
                    if exc.errno == ENOTSOCK:
                        _log.debug("Socket send on non-socket %r", self.core().identity)
        return results or None

    def call(self, peer, method, *args, **kwargs):
        if self._protected_rpcs is None and peer not in [AUTH, CONTROL_CONNECTION, CONTROL]:
            # first rpc call
            # TODO: can this be done on onconnect? or on some other event
            self._protected_rpcs = self.get_protected_rpcs()
            self._wrap_protected_rpcs(self._protected_rpcs)
        self_ref = kwargs.pop("self", None)
        platform = kwargs.pop("external_platform", "")
        request, result = self._dispatcher.call(method, args, kwargs)
        ident = f"{next(self._counter)}.{hash(result)}"
        self._outstanding[ident] = result
        subsystem = None
        frames = []

        if not self._isconnected:
            return

        if platform == "":    # local platform
            subsystem = "RPC"
            frames.append(request)
        else:
            frames = []
            operation = "send_platform"
            subsystem = "external_rpc"
            frames.append(operation)
            msg = dict(
                to_platform=platform,
                to_peer=peer,
                from_platform="",
                from_peer="",
                args=[request],
            )
            frames.append(msg)
            peer = ""

        try:
            self.core().connection.send_vip(peer, subsystem, args=frames, msg_id=ident)
        except ZMQError as exc:
            if exc.errno == ENOTSOCK:
                _log.debug("Socket send on non-socket %r", self.core().identity)

        return result

    def get_protected_rpcs(self):
        return self.call(AUTH, "get_protected_rpcs", self._owner.core.identity).get(timeout=10)

    __call__ = call

    def notify(self, peer, method, *args, **kwargs):
        platform = kwargs.pop("external_platform", "")
        request = self._dispatcher.notify(method, args, kwargs)
        frames = []
        if not self._isconnected:
            return

        subsystem = None
        if platform == "":
            subsystem = "RPC"
            frames.append(request)
        else:
            operation = "send_platform"
            subsystem = "external_rpc"
            frames.append(operation)
            msg = dict(
                to_platform=platform,
                to_peer=peer,
                from_platform="",
                from_peer="",
                args=[request],
            )
            frames.append(msg)
            peer = ""

        try:
            self.core().connection.send_vip(peer, subsystem, args=frames)
        except ZMQError as exc:
            if exc.errno == ENOTSOCK:
                _log.debug("Socket send on non-socket %r", self.core().identity)



import argparse
import json
import re

from typing import Callable, List

from volttron.utils.context import ClientContext as cc

import volttron.types.auth.authz_types as authz

# Note: rpc call in volttron-lib-auth/src/volttron/services/auth/auth_service.py
from volttron.services.auth.auth_service import AUTH, VolttronAuthService
from volttron.utils.jsonrpc import RemoteError


def add_authz_parser(add_parser_fn, filterable):
    """Create and populate an argparse parser for the authz command.

    First create the top level parser for authz.  Then create a subparser for
    the rpc subcommand.  Finally adds separate arguments to the rpc subparser
    for add, remove and list commands.

    The same method as above is how the pubsub subcommand will be added.

    :param add_parser_fn: A callback that will create a new parser based upon parameters passed
    :type add_parser_fn: Callable
    :param filterable: A filter function for filtering the results of the command
    :type filterable: Callable

    EXAMPLE:
    vctl authz -h
    usage: vctl command [OPTIONS] ... authz [-h] [--debug] [-t SECS] [--address ADDR]  ...

    options:
    -h, --help            show this help message and exit
    --debug               show tracebacks for errors rather than a brief message
    -t SECS, --timeout SECS
                            timeout in seconds for remote calls (default: 60)
    --address ADDR        URL to bind for VIP connections

    authz operations:

        add                 Add rpc method authorization
        remove              Remove rpc method authorization
        list                List authorized rpc methods.
    """

    # TODO: Verify that the filterable makes sense for the authz command.

    authz_command = add_parser_fn("authz", help="Manage authorization for rpc methods and pubsub topics")

    authz_operations = authz_command.add_subparsers(
        title="Authorization operations",
        metavar="<operations=add|remove|list>",
        dest="store_commands",
        required=True,
        help="Operations to add/merge, remove, and list volttron authorization rules.",
    )

    # Create the 'add' subparser under 'rpc'
    authz_add_parser = authz_operations.add_parser(
        "add", help="Add or merge a role, group,  agent authorization, or protected topics")
    # authz_add_parser.add_argument("identity_and_method", nargs="*", help="Format: 'identity.method_name'")
    # authz_add_parser.set_defaults(func=handel_role_parser)

    #### Create subparser for node ('role', 'group', 'topics', 'agent') under 'authz add'
    add_node_parser = authz_add_parser.add_subparsers(
        title="Add or merge role, group, agent authorization or protected topic entry",
        metavar="<NODE=role|group|topics|agent>",
        dest="store_commands",
        required=True,
    )
    capabilities_epilog = f"""
Additional information:\n
pubsub-capabilities:
    {AuthZUtils.topic_pattern_pubsub_constrain_valid_requirement()}
rpc-capabilities:
    {AuthZUtils.rpc_capability_format_requirement()}"""
    # Add a command "role" under 'authz add'
    add_role_command = add_node_parser.add_parser("role",
                                                  help="create a role or merge rules for an existing role",
                                                  formatter_class=argparse.RawDescriptionHelpFormatter,
                                                  epilog=capabilities_epilog)
    add_role_command.add_argument("role_name", help="name of the role")
    add_role_command.add_argument(
        "--pubsub-capabilities",
        "-ps",
        nargs="+",
        help="one or more pubsub-capabilities allowed for this role. see additional information")
    add_role_command.add_argument("--rpc-capabilities",
                                  "-rpc",
                                  nargs="+",
                                  help="one or more rpc capabilities allowed for this role. see additional information")
    add_role_command.set_defaults(func=authz_add_role)

    # Add a command "group" under 'authz add'
    add_group_command = add_node_parser.add_parser("group",
                                                   help="create a group or merge rules for an existing group",
                                                   formatter_class=argparse.RawDescriptionHelpFormatter,
                                                   epilog=capabilities_epilog)
    add_group_command.add_argument("group_name", help="name of the group")
    add_group_command.add_argument(
        "identities", nargs="+",
        help="one or more agent identity to be added to group")    # "+" means one or more inputs are required,
    add_group_command.add_argument("--role-names", "-rns", nargs="+", help="name of roles to be assigned to this group")
    add_group_command.add_argument(
        "--pubsub-capabilities",
        "-ps",
        nargs="+",
        help="one or more pubsub-capabilities allowed for this group. see additional information",
    )
    add_group_command.add_argument(
        "--rpc-capabilities",
        "-rpc",
        nargs="+",
        help="one or more rpc-capabilities allowed for this role. see additional information")
    add_group_command.set_defaults(func=authz_add_group)

    # Add a command "topics" under 'authz add'
    add_topic_command = add_node_parser.add_parser(
        "topics",
        help=("protect one or more topic or topic pattern by adding it to the protected_topics list."
              "topic patterns(regular expression) should be enclose within // Example: /devices.*/. "),
    )
    add_topic_command.add_argument("topic_names",
                                   nargs="+",
                                   help="one or more topic name or topic pattern to be protected")
    add_topic_command.set_defaults(func=authz_add_topic)

    # Add a command "agent" under 'authz add'
    add_agent_command = add_node_parser.add_parser("agent",
                                                   help="create or merge an agent's authorization rules",
                                                   formatter_class=argparse.RawDescriptionHelpFormatter,
                                                   epilog=capabilities_epilog)
    add_agent_command.add_argument("identity", help="vip identity of the agent")
    add_agent_command.add_argument("--role-names",
                                   "-rns",
                                   nargs="+",
                                   help="name of role(s) to be assigned to this agent")
    add_agent_command.add_argument("--protected-rpcs",
                                   "-pts",
                                   nargs="+",
                                   help="rpc exported methods of this agent that needs to be protected by "
                                   "authorization rules")
    add_agent_command.add_argument(
        "--pubsub-capabilities",
        "-ps",
        nargs="+",
        help="one or more pubsub-capabilities allowed for this group. see additional information",
    )
    add_agent_command.add_argument(
        "--rpc-capabilities",
        "-rpc",
        nargs="+",
        help="one or more rpc-capabilities allowed for this group. see additional information")
    add_agent_command.add_argument("--comments", "-c", help="comment string")
    add_agent_command.set_defaults(func=authz_add_agent)

    ### REMOVE parser
    remove_authz_method = add_parser_fn(
        "remove",
        subparser=authz_operations,
        help="Remove authorization entries for a agent, group, role or protected topics")

    #### Create subparser for node ('role', 'group', 'topics', 'agent') under 'authz remove'
    remove_node_parser = remove_authz_method.add_subparsers(
        title="top nodes",
        metavar="<NODE=role|group|topics|agent>",
        dest="store_commands",
        required=True,
    )

    # Add a command "role" under 'authz remove'
    remove_role_command = remove_node_parser.add_parser("role", help="remove a given role")
    remove_role_command.add_argument("role_name", help="name of the role to be removed")
    remove_role_command.set_defaults(func=authz_remove_role)

    # Add a command "group" under 'authz remove'
    remove_group_command = remove_node_parser.add_parser("group", help="remove a given group")
    remove_group_command.add_argument("group_name", help="name of the group to be removed")
    remove_group_command.set_defaults(func=authz_remove_group)

    # Add a command "topics" under 'authz remove'
    remove_topic_command = remove_node_parser.add_parser(
        "topics", help="remove one or more topic or pattern from protected topics list")
    remove_topic_command.add_argument("topic_names",
                                      nargs="+",
                                      help="space separated list of topic or topic pattern to remove")
    remove_topic_command.set_defaults(func=authz_remove_topic)

    # Add a command "agent" under 'authz remove'
    remove_agent_command = remove_node_parser.add_parser("agent", help="remove an agent authorization rules")
    remove_agent_command.add_argument("identity",
                                      help="vip identity of the agent")    # "+" means one or more inputs are required,
    remove_agent_command.set_defaults(func=authz_remove_agent)

    ### LIST parser
    list_authz_method = add_parser_fn("list", subparser=authz_operations, help="List authorized rpc methods.")
    list_node_parser = list_authz_method.add_subparsers(
        title="top nodes",
        metavar="<NODE=role|group|topics|agent>",
        dest="store_commands",
        required=False,
    )
    # Add a command "role" under 'authz remove'
    list_role_command = list_node_parser.add_parser("role", help="list role")
    # Add a command "group" under 'authz remove'
    list_group_command = list_node_parser.add_parser("group", help="list group")
    # Add a command "protected-topics" under 'authz remove'
    list_topic_command = list_node_parser.add_parser("topics", help="list protected topics")
    # Add a command "agent" under 'authz remove'
    list_agent_command = list_node_parser.add_parser("agent", help="list agent")
    # list_authz_method.set_defaults(func=list_dummy)
    list_authz_method.set_defaults(func=authz_list_dummy)

    list_role_command.set_defaults(func=authz_list_dummy)
    list_group_command.set_defaults(func=authz_list_dummy)
    list_topic_command.set_defaults(func=authz_list_dummy)
    list_agent_command.set_defaults(func=authz_list_dummy)


# def print_args(opts):
#     return f"{opts=}"


def get_local_authorizations() -> dict[str, dict]:
    # TODO maybe cache this maybe not depending, but we have made it so that only one place deals with io.
    with open(cc.get_volttron_home().join("/authz.json")) as fp:
        return json.load(fp)


def list_dummy(opts):
    """TODO: clean-up: Mock method for `vctl authz list` for now"""
    data = get_local_authorizations()
    # print(json.dumps(data, indent=4))
    print(data.keys())
    return data


def authz_list_dummy(opts):
    """TODO: clean-up: Mock method for `vctl authz list <node>` for now"""
    data = get_local_authorizations()
    list_content = data
    # opts.store_commands in ["role", "group", "topic", "agent"]:  #
    if opts.store_commands == "role":
        list_content = data.get("roles")
    elif opts.store_commands == "agent":
        list_content = data.get("agents")
    elif opts.store_commands == "group":
        list_content = data.get("agent_groups")
    elif opts.store_commands == "topics":
        list_content = data.get("protected_topics")
    else:
        pass
    # return list_content
    print(json.dumps(list_content, indent=4))
    # print(list_content.keys())


### authz control
def authz_add_role(opts):
    role_name: str = opts.role_name
    rpc_capabilities_attr: List[str] | None = opts.rpc_capabilities
    pubsub_capabilities_attr: List[str] | None = opts.pubsub_capabilities

    rpc_capabilities = AuthZUtils.str_to_RPCCapabilities(rpc_capabilities_attr)
    pubsub_capabilities = AuthZUtils.str_to_PubsubCapabilities(pubsub_capabilities_attr)

    rpc_method: Callable = VolttronAuthService.create_or_merge_role
    res = opts.connection.server.vip.rpc.call(AUTH,
                                              rpc_method.__name__,
                                              name=role_name,
                                              pubsub_capabilities=pubsub_capabilities,
                                              rpc_capabilities=rpc_capabilities).get()
    if res:
        print(f"Added Role: {rpc_capabilities_attr=}, {pubsub_capabilities_attr=} to {role_name=}.")


def authz_remove_role(opts):
    role_name: str = opts.role_name
    rpc_method: Callable = VolttronAuthService.remove_role
    res = opts.connection.server.vip.rpc.call(AUTH, rpc_method.__name__, name=role_name).get()
    if res:
        print(f"Removed Role: {role_name=}.")


def authz_add_agent(opts):
    identity: str = opts.identity
    role_names: List[str] | None = opts.role_names
    protected_rpcs: List[str] | None = opts.protected_rpcs
    rpc_capabilities_attr: List[str] | None = opts.rpc_capabilities
    pubsub_capabilities_attr: List[str] | None = opts.pubsub_capabilities
    comments: str | None = opts.comments

    rpc_capabilities = AuthZUtils.str_to_RPCCapabilities(rpc_capabilities_attr)
    pubsub_capabilities = AuthZUtils.str_to_PubsubCapabilities(pubsub_capabilities_attr)
    roles = AuthZUtils.str_to_AgentRoles(role_names)

    rpc_method: Callable = VolttronAuthService.create_or_merge_agent_authz
    res = opts.connection.server.vip.rpc.call(
        AUTH,
        rpc_method.__name__,
        identity=identity,
        protected_rpcs=protected_rpcs,
        roles=roles,
        pubsub_capabilities=pubsub_capabilities,
        rpc_capabilities=rpc_capabilities,
        comments=comments,
    ).get()
    if res:
        print(f"Added Agent:  {role_names=}, \
{rpc_capabilities_attr=}, {pubsub_capabilities_attr=}, \
{comments=} to {identity=}.")


def authz_remove_agent(opts):
    identity: str = opts.identity
    rpc_method: Callable = VolttronAuthService.remove_agent
    # TODO: remove_agent is not robust. Often got "volttron.utils.jsonrpc.RemoteError: volttron.types.auth.auth_credentials.IdentityNotFound('role7')" need to figure out why.
    res = opts.connection.server.vip.rpc.call(
        AUTH,
        rpc_method.__name__,
        identity=identity,
    ).get()
    if res:
        print(f"Removed Agent: {identity=}.")


def authz_add_topic(opts):
    topic_names: List[str] = opts.topic_names
    protected_topics = AuthZUtils.str_to_topic_patterns(topic_names)
    rpc_method: Callable = VolttronAuthService.create_protected_topics
    res = opts.connection.server.vip.rpc.call(
        AUTH,
        rpc_method.__name__,
        topic_name_patterns=protected_topics,
    ).get()
    if res:
        print(f"Added Topic: {topic_names=}.")


def authz_remove_topic(opts):
    topic_names: str = opts.topic_names
    rpc_method: Callable = VolttronAuthService.remove_protected_topics
    res = opts.connection.server.vip.rpc.call(
        AUTH,
        rpc_method.__name__,
        topic_name_patterns=topic_names,
    ).get()
    if res:
        print(f"Removed Topic: {topic_names=}.")
    else:
        # TODO error with silence if no such topic exists.
        print(f"SOMEHTING WRONG, probabely, one of such topics {topic_names=} didn't exist.")


def authz_add_group(opts):
    group_name: str = opts.group_name
    vip_ids: List[str] = opts.identities
    role_names: List[str] | None = opts.role_names
    # topic_names: List[str] | None = opts.topic_names
    rpc_capabilities_attr: List[str] | None = opts.rpc_capabilities
    pubsub_capabilities_attr: List[str] | None = opts.pubsub_capabilities

    if not any([role_names, rpc_capabilities_attr, pubsub_capabilities_attr]):    # TODO: should we handle this here?
        raise ValueError(
            "agent group group1 should have non empty capabilities. Please pass non empty values for at least one of the three parameters - agent_roles, rpc_capabilities, pubsub_capabilities"
        )

    rpc_capabilities = AuthZUtils.str_to_RPCCapabilities(rpc_capabilities_attr)
    pubsub_capabilities = AuthZUtils.str_to_PubsubCapabilities(pubsub_capabilities_attr)
    roles = AuthZUtils.str_to_AgentRoles(role_names)

    rpc_method: Callable = VolttronAuthService.create_or_merge_agent_group
    res = opts.connection.server.vip.rpc.call(
        AUTH,    # "platform.auth",
        rpc_method.__name__,
        name=group_name,
        identities=vip_ids,
    # protected_rpcs=protected_rpcs,
        roles=roles,
        pubsub_capabilities=pubsub_capabilities,
        rpc_capabilities=rpc_capabilities,
    ).get()
    if res:
        print(f"Added Group: {role_names=}, \
{rpc_capabilities_attr=}, {pubsub_capabilities_attr=}, \
to {group_name=}.")


def authz_remove_group(opts):
    group_name: str = opts.group_name
    rpc_method: Callable = VolttronAuthService.remove_agent_group
    res = opts.connection.server.vip.rpc.call(
        AUTH,
        rpc_method.__name__,
        name=group_name,
    ).get()
    if res:
        print(f"Removed Group: {group_name=}.")
    else:
        # TODO error with silence if no such topic exists.
        print(f"SOMEHTING WRONG, probabely, one of such groups {group_name=} didn't exist.")


class AuthZUtils:

    @staticmethod
    def is_rpc_capability_format_valid(cap_attr: str) -> bool:
        """
        Validates that the value follows the 'string.string' format.
        This function uses regular expression to check the pattern.
        """
        # patern vipid.rpcmethod
        # vip id could have . in it. Ex. config.store.edit_config
        pattern = re.compile(r"^\w+[.\w]*\.\w+$")
        return bool(pattern.match(cap_attr))

    @staticmethod
    def rpc_capability_format_requirement() -> str:
        return "rpc-capabilities should be of the format '<vip id>.<rpc method name>' format. i.e., 'agent_1.rpcmethod1'"

    @staticmethod
    def is_topic_pattern_valid(topic_pattern: str) -> bool:
        """
        Check if the provided string matches the specific pattern:
        Can contain letters, '/', '.', '*', brackets, hyphens, undercore, and plus signs.

        Args:
        s (str): The string to be checked.

        Returns:
        bool: True if the string matches the format, False otherwise.

        # Examples of usage:
        test_strings = [
            "/devicez/ahu.*/",     # valid: follows specified characters and pattern
            "/devicez/ahu[1-9]+/", # valid: includes numbers and regex patterns
            "/devicez/ahu-123*/",  # valid: hyphen and asterisk used correctly
            "/devicez/ahu+/",      # valid: plus sign used correctly
            "/*/auth.*/",          # valid: asterisk used at the beginning and in pattern
            "/invalid_string$/",   # invalid: dollar sign is not in the allowed set
            "/devicez/ahu(!)/",    # invalid: parentheses are not allowed
            "/devicez|ahu.*/",     # invalid: pipe character is not allowed
            "/devicez/ahu[1-9]*/", # valid: correct use of brackets and asterisk
            "devicez/ahu{}",     # invalid: curly brackets are not allowed
            "hello world"        # invalid: space is not allowed
        ]
        """
        # Regex pattern to match the specified format
        pattern = r"^[a-zA-Z0-9/\.\*\[\]\-\+\_]*$"

        # Check if the string matches the pattern
        if re.match(pattern, topic_pattern):
            return True
        else:
            return False

    @staticmethod
    def topic_pattern_requirement() -> str:
        example_usage = r"""
        test_strings = [
            "/devicez/ahu.*/",     # valid: follows specified characters and pattern
            "/devicez/ahu[1-9]+/", # valid: includes numbers and regex patterns
            "/devicez/ahu-123*/",  # valid: hyphen and asterisk used correctly
            "/devicez/ahu+/",      # valid: plus sign used correctly
            "/*/auth.*/",          # valid: asterisk used at the beginning and in pattern
            "/invalid_string$/",   # invalid: dollar sign is not in the allowed set
            "/devicez/ahu(!)/",    # invalid: parentheses are not allowed
            "/devicez|ahu.*/",     # invalid: pipe character is not allowed
            "/devicez/ahu[1-9]*/", # valid: correct use of brackets and asterisk
            "devicez/ahu{}",     # invalid: curly brackets are not allowed
            "hello world"        # invalid: space is not allowed
        ]"""
        # return f"Can contain letters, '/', '.', '*', brackets, hyphens, and plus signs. {example_usage=}"
        return bytes(
            f"Can contain letters, '/', '.', '*', brackets, hyphens, and plus signs. {example_usage=}",
            "utf-8",
        ).decode("unicode_escape")    # Manually interpreting escape sequences

    @staticmethod
    def is_pubsub_constrain_valid(pubsub_constrain: str) -> bool:
        """
        Validates pubsub_constrain
        """
        # return pubsub_constrain in ["publish", "subscribe", "pub", "sub", "pubsub"]
        return pubsub_constrain in ["publish", "subscribe", "pubsub"]

    @staticmethod
    def pubsub_constrain_requirement() -> str:
        return 'topic_access in ["publish", "subscribe", "pubsub"]'

    @classmethod
    def is_topic_pattern_pubsub_constrain_valid(cls, input_string: str) -> bool:
        """
        Checks if the input string follows the format '<topic_pattern>:<pubsub_constraint>'

        Args:
        input_string (str): The input string to validate.

        Returns:
        bool: True if the input string is valid, False otherwise.
        """
        # Split the input string by the colon
        parts = input_string.split(":")

        # Ensure there are exactly two parts
        if len(parts) != 2:
            return False

        # Validate each part
        topic_pattern, pubsub_constrain = parts
        return cls.is_topic_pattern_valid(topic_pattern) and cls.is_pubsub_constrain_valid(pubsub_constrain)

    @staticmethod
    def topic_pattern_pubsub_constrain_valid_requirement() -> str:
        example_usage = r"""
        [
            "/devicez/ahu.*/:publish",         # valid
            "/auth.*/:pubsub",               # valid
            "devicez/ahu(!):publish",        # invalid: topic pattern invalid
            "devicez|ahu.*:fly",             # invalid: pubsub constraint invalid
            "/prefix/literal-topic-match:subscribe"    # valid
        ]"""
        # print(
        #     "The input string needs to follow the format '<topic_pattern>:<pubsub_constraint>'."
        # )
        # print(example_usage)
        # return f"The input string needs to follow the format '<topic_pattern>:<pubsub_constraint>'. {example_usage=}"
        return bytes(
            f"pubsub-capabilities should be of format '<topic>:<publish|subscribe|pubsub>'. If topic is a "
            f"regular expression enclosed within //\n"
            f"{example_usage}",
            "utf-8",
        ).decode("unicode_escape")    # Manually interpreting escape sequences

    # TODO: this method (str_to_RPCCapabilities and str_to_xx methods) should be adopted by authz.RPCCapabilities itself.
    @staticmethod
    def str_to_RPCCapabilities(rpc_capabilities_attr: List[str] | None, ) -> authz.RPCCapabilities | None:
        if rpc_capabilities_attr is None:
            return None
        # check rpc_cap in "id.rpc1" format
        rpc_caps: List[authz.RPCCapability] = []
        for rpc_cap in rpc_capabilities_attr:
            if not AuthZUtils.is_rpc_capability_format_valid(rpc_cap):
                msg = f"Input rpc-capability '{rpc_cap}' in {rpc_capabilities_attr} does not meet the required format: {AuthZUtils.rpc_capability_format_requirement()}"
                raise ValueError(msg)
            rpc_caps.append(authz.RPCCapability(rpc_cap))

        return authz.RPCCapabilities(rpc_caps)

    @staticmethod
    def str_to_PubsubCapabilities(pubsub_capabilities_attr: List[str] | None) -> authz.PubsubCapabilities | None:
        if pubsub_capabilities_attr is None:
            return None
        pubsub_caps = []
        # check pubsub_cap in "devicez/ahu.*:publish" format
        for pubsub_cap in pubsub_capabilities_attr:
            if ":" not in pubsub_cap:
                msg = f"Input pubsub-capability '{pubsub_cap}' in {pubsub_capabilities_attr} does not meet the required format: {AuthZUtils.topic_pattern_pubsub_constrain_valid_requirement()}"
                raise ValueError(msg)
            topic_pattern = pubsub_cap.split(":")[0]
            topic_access = pubsub_cap.split(":")[-1]
            if not AuthZUtils.is_topic_pattern_valid(topic_pattern):
                raise ValueError(
                    f"Input '<{topic_pattern=}>:<pubsub_constraint>' in {pubsub_capabilities_attr} does not meet the required format: {AuthZUtils.topic_pattern_requirement()}"
                )
            if not AuthZUtils.is_pubsub_constrain_valid(topic_access):
                raise ValueError(
                    f"Input '<topic_pattern>:<{topic_access=}>:' in {pubsub_capabilities_attr} does not meet the required format: {AuthZUtils.pubsub_constrain_requirement()}"
                )
            pubsub_caps.append(authz.PubsubCapability(topic_pattern=topic_pattern, topic_access=topic_access))
        return authz.PubsubCapabilities(pubsub_caps)

    @staticmethod
    def str_to_topic_patterns(topic_names: List[str] | None, ) -> List[str] | None:
        if topic_names is None:
            return None
        protected_rpcs: List[str] = []
        for topic_name in topic_names:
            if not AuthZUtils.is_topic_pattern_valid(topic_name):
                raise ValueError(
                    f"Input '{topic_name=}' in {topic_names} does not meet the required format: {AuthZUtils.topic_pattern_requirement()}"
                )
            protected_rpcs.append(topic_name)
        return protected_rpcs

    @staticmethod
    def str_to_AgentRoles(role_names: List[str] | None) -> authz.AgentRoles | None:
        if role_names is None:
            return None
        roles = authz.AgentRoles([authz.AgentRole(role_name=role_name) for role_name in role_names])
        return roles



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

__all__ = [
    "execute_command", "vip_main", "is_volttron_running", "wait_for_volttron_startup",
    "wait_for_volttron_shutdown", "start_agent_thread", "isapipe"
]

import logging
import os
from pathlib import Path
import subprocess
import stat
import sys

import gevent
import psutil

_log = logging.getLogger(__name__)


def execute_command(cmds, env=None, cwd=None, logger=None, err_prefix=None) -> str:
    """Executes a command as a subprocess (Not greenlet safe!)

    If the return code of the call is 0 then return stdout otherwise
    raise a RuntimeError.  If logger is specified then write the exception
    to the logger otherwise this call will remain silent.

    :param cmds:list of commands to pass to subprocess.run
    :param env: environment to run the command with
    :param cwd: working directory for the command
    :param logger: a logger to use if errors occur
    :param err_prefix: an error prefix to allow better tracing through the error message
    :return: stdout string if successful

    :raises RuntimeError: if the return code is not 0 from suprocess.run
    """

    results = subprocess.run(cmds,
                             env=env,
                             cwd=cwd,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
    if results.returncode != 0:
        err_prefix = err_prefix if err_prefix is not None else "Error executing command"
        err_message = ("\n{}: Below Command failed with non zero exit code.\n"
                       "Command:{} \nStdout:\n{}\nStderr:\n{}\n".format(
                           err_prefix, results.args, results.stdout, results.stderr))
        if logger:
            logger.exception(err_message)
            raise RuntimeError()
        else:
            raise RuntimeError(err_message)

    return results.stdout.decode("utf-8")


def start_agent_thread(cls, **kwargs):
    """Instantiate an agent class and run it in a new daemon thread.

    Returns the thread object.
    """
    import threading

    agent = cls(**kwargs)
    thread = threading.Thread(target=agent.run)
    thread.daemon = True
    thread.start()
    return thread


def isapipe(fd):
    fd = getattr(fd, "fileno", lambda: fd)()
    return stat.S_ISFIFO(os.fstat(fd).st_mode)


def vip_main(agent_class, version: str = "0.1", **kwargs):
    """Default main entry point implementation for VIP agents."""
    from volttron.client.vip.agent import Agent
    from volttron.utils import (ClientContext as cc, is_valid_identity, get_address)
    from volttron.types.auth.auth_credentials import CredentialsFactory
    try:
        # If stdout is a pipe, re-open it line buffered
        if isapipe(sys.stdout):
            # Hold a reference to the previous file object so it doesn't
            # get garbage collected and close the underlying descriptor.
            stdout = sys.stdout
            sys.stdout = os.fdopen(stdout.fileno(), "w", 1)

        # Quiet printing of KeyboardInterrupt by greenlets
        Hub = gevent.hub.Hub
        Hub.NOT_ERROR = Hub.NOT_ERROR + (KeyboardInterrupt, )

        config = os.environ.pop("AGENT_CONFIG", {})
        identity = os.environ.pop("AGENT_VIP_IDENTITY", None)
        address = os.environ.pop("VOLTTRON_PLATFORM_ADDRESS", None)

        creds = CredentialsFactory.load_from_environ()
        del os.environ["AGENT_CREDENTIALS"]

        if not os.environ.get("_LAUNCHED_BY_PLATFORM"):
            if not creds:
                raise ValueError(
                    "Please set AGENT_CREDENTIALS to the credentials to connect to the server.")
            if not address:
                raise ValueError(
                    "Please set VOLTTRON_PLATFORM_ADDRESS to the address for connecting to the server."
                )

        if identity != creds.identity:
            raise ValueError("AGENT_VIP_IDENTITY and identity from credentials do not match!")

        if identity is not None:
            if not is_valid_identity(identity):
                _log.warning("Deprecation warning")
                _log.warning(f"All characters in {identity} are not in the valid set.")

        agent = agent_class(credentials=creds, config_path=config, address=address, **kwargs)

        try:
            run = agent.run
        except AttributeError:
            run = agent.core.run
        task = gevent.spawn(run)
        try:
            task.join()
        finally:
            task.kill()
    except KeyboardInterrupt:
        pass


def is_volttron_running(volttron_home):
    """
    Checks if volttron is running for the given volttron home. Checks if a VOLTTRON_PID file exist and if it does
    check if the PID in the file corresponds to a running process. If so, returns True else returns False
    :param vhome: volttron home
    :return: True if VOLTTRON_PID file exists and points to a valid process id
    """

    pid_file = os.path.join(volttron_home, "VOLTTRON_PID")
    if os.path.exists(pid_file):
        running = False
        with open(pid_file, "r") as pf:
            pid = int(pf.read().strip())
            running = psutil.pid_exists(pid)
        return running
    else:
        return False


def wait_for_volttron_startup(vhome, timeout):
    # Check for VOLTTRON_PID
    sleep_time = 0
    while (not is_volttron_running(vhome)) and sleep_time < timeout:
        gevent.sleep(3)
        sleep_time += 3
    if sleep_time >= timeout:
        raise Exception("Platform startup failed. Please check volttron.log in {}".format(vhome))


def wait_for_volttron_shutdown(vhome, timeout):
    # Check for VOLTTRON_PID
    sleep_time = 0
    while (is_volttron_running(vhome)) and sleep_time < timeout:
        gevent.sleep(1)
        sleep_time += 1
    if sleep_time >= timeout:
        raise Exception(
            "Platform shutdown failed. Please check volttron.cfg.log in {}".format(vhome))



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

__all__ = ["is_valid_identity", "normalize_identity"]

import re

# The following are the only allowable characters for identities.
_VALID_IDENTITY_RE = re.compile(r"^[A-Za-z0-9_.-]+$")


def is_valid_identity(identity_to_check):
    """Checks the passed identity to see if it contains invalid characters

    A None value for identity_to_check will return False

    @:param: string: The vip_identity to check for validity
    @:return: boolean: True if values are in the set of valid characters.
    """

    if identity_to_check is None:
        return False

    return _VALID_IDENTITY_RE.match(identity_to_check)


def normalize_identity(pre_identity):
    if is_valid_identity(pre_identity):
        return pre_identity

    if pre_identity is None:
        raise ValueError("Identity cannot be none.")

    norm = ""
    for s in pre_identity:
        if _VALID_IDENTITY_RE.match(s):
            norm += s
        else:
            norm += "_"

    return norm



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

__all__ = ["get_version"]

from pathlib import Path
import importlib.metadata as importlib_metadata
import sys

from volttron.utils import execute_command

# Version is going to start out as None and get set when import is done.
__version__ = None

# Try to get the version from written metadata, but
# if failed then get it from the pyproject.toml file
try:
    # Note this is the wheel prefix or the name attribute in pyproject.toml file.
    # this is the version of the program that is used when the application is installed
    # via a wheel.
    __version__ = importlib_metadata.version('volttron-core')

except importlib_metadata.PackageNotFoundError:
    try:
        # We should be in a develop environment therefore
        # we can get the version from the toml pyproject.toml
        root = Path(__file__).parent.parent.parent
        tomle_file = root.joinpath("pyproject.toml")
        if not tomle_file.exists():
            raise ValueError(
                f"Couldn't find pyproject.toml file for finding version. ({str(tomle_file)})")
        import toml

        pyproject = toml.load(tomle_file)

        __version__ = pyproject["tool"]["poetry"]["version"]
    except ValueError:
        cmd = [sys.executable, "-m", "pip", "list"]
        results = execute_command(cmd).split("\n")
        for r in results:
            if r.find("volttron-core") >= 0:
                # split trims and removes whitespace around the elements.
                __version__ = r.split()[1]
                break

if not __version__:
    raise ImportError("Couldn't find version for volttron-core library!")


def get_version():
    """
    Return the version number of the platform.  This function handles both cases where
    we are in developer mode (i.e. there is a pyproject.toml file) and when it is
    installed in a deployed environment where the version is looked up from the parent.
    """
    return __version__



from typing import Any, Optional
from abc import ABC, abstractmethod
from typing import Literal

from volttron.types.auth.auth_credentials import (Credentials, CredentialsCreator, CredentialsStore)

from volttron.types import Service, Identity

import volttron.types.auth.authz_types as authz


# TODO Make AuthorizationManager the Authorizer.
#  Below is not used
class Authorizer(ABC):
    pass
    # @abstractmethod
    # def check_rpc_authorization(self, *, identity: authz.Identity, method_name: authz.vipid_dot_rpc_method,
    #                             method_args: dict, **kwargs) -> bool:
    #     ...
    #
    # @abstractmethod
    # def check_pubsub_authorization(self, *, identity: authz.Identity,
    #                                topic_pattern: str, access: str, **kwargs) -> bool:
    #     ...


class AuthzPersistence(ABC):

    @classmethod
    def load(cls, authz_input: Any, **kwargs) -> authz.VolttronAuthzMap:
        ...

    @classmethod
    def store(cls, authz_map: authz.VolttronAuthzMap, **kwargs) -> bool:
        ...


class Authenticator(ABC):

    @abstractmethod
    def is_authenticated(self, *, identity: authz.Identity) -> bool:
        ...


class AuthorizationManager:

    @abstractmethod
    def get_protected_rpcs(self, identity: authz.Identity) -> list[str]:
        ...

    @abstractmethod
    def check_rpc_authorization(self, *, identity: authz.Identity, method_name: authz.vipid_dot_rpc_method,
                                method_args: dict, **kwargs) -> bool:
        ...

    @abstractmethod
    def check_pubsub_authorization(self, *, identity: authz.Identity, topic_pattern: str, access: str,
                                   **kwargs) -> bool:
        ...

    @abstractmethod
    def create_or_merge_role(self,
                             *,
                             name: str,
                             rpc_capabilities: Optional[authz.RPCCapabilities] = None,
                             pubsub_capabilities: Optional[authz.PubsubCapabilities] = None,
                             **kwargs) -> bool:
        ...

    @abstractmethod
    def create_or_merge_agent_group(self,
                                    *,
                                    name: str,
                                    identities: list[authz.Identity],
                                    roles: Optional[authz.AgentRoles] = None,
                                    rpc_capabilities: Optional[authz.RPCCapabilities] = None,
                                    pubsub_capabilities: Optional[authz.PubsubCapabilities] = None,
                                    **kwargs) -> bool:
        ...

    @abstractmethod
    def remove_agents_from_group(self, name: str, identities: list[authz.Identity]):
        ...

    @abstractmethod
    def add_agents_to_group(self, name: str, identities: list[authz.Identity]):
        ...

    @abstractmethod
    def create_or_merge_agent_authz(self,
                                    *,
                                    identity: str,
                                    protected_rpcs: list[str] = None,
                                    roles: Optional[authz.AgentRoles] = None,
                                    rpc_capabilities: Optional[authz.RPCCapabilities] = None,
                                    pubsub_capabilities: Optional[authz.PubsubCapabilities] = None,
                                    comments: str = None,
                                    **kwargs) -> bool:
        ...

    @abstractmethod
    def get_agent_capabilities(self, *, identity: str) -> dict:
        ...

    @abstractmethod
    def create_protected_topics(self, *, topic_name_patterns: list[str]) -> bool:
        ...

    @abstractmethod
    def is_protected_topic(self, *, topic_name_pattern: str) -> bool:
        """Return True if the topic is protected, False otherwise.

        The topic_expression can be a str or regex pattern.  If the string or
        expression matches a protected topic then True is returned.  If not,
        False is returned.

        :param topic_name_pattern: The topic to check if it is protected.
        :type topic_name_pattern: str
        :return: True if the topic is protected, False otherwise.
        :rtype: bool
        """
        ...

    @abstractmethod
    def remove_protected_topics(self, *, topic_name_patterns: list[str]) -> bool:
        ...

    @abstractmethod
    def remove_agent_authorization(self, identity: authz.Identity):
        ...

    @abstractmethod
    def remove_agent_group(self, name: str):
        ...

    @abstractmethod
    def remove_role(self, name: str):
        ...


class AuthService(Service):

    # Authentication
    @abstractmethod
    def create_agent(self, *, identity: str, **kwargs) -> bool:
        ...

    @abstractmethod
    def remove_agent(self, *, identity: str, **kwargs) -> bool:
        ...

    @abstractmethod
    def has_credentials_for(self, *, identity: Identity) -> bool:
        ...

    @abstractmethod
    def create_credentials(self, *, identity: Identity):
        ...

    @abstractmethod
    def remove_credentials(self, *, identity: Identity):
        ...

    @abstractmethod
    def is_credentials(self, *, identity: Identity) -> bool:
        ...

    # Authorization

    @abstractmethod
    def get_protected_rpcs(self, identity: authz.Identity) -> list[str]:
        """
        returns list of protected methods for a given identity
        """
        ...

    @abstractmethod
    def check_rpc_authorization(self, *, identity: authz.Identity, method_name: authz.vipid_dot_rpc_method,
                                method_args: dict, **kwargs) -> bool:
        """ should throw AuthException is calling agent(identity) is not authorized to access the
            method_name(vip_id.rpc_method) with the specific arguments method_args"""
        ...

    @abstractmethod
    def check_pubsub_authorization(self, *, identity: authz.Identity, topic_pattern: str,
                                   access: Literal["pubsub", "publish", "subscribe"], **kwargs) -> bool:
        ...

    @abstractmethod
    def create_or_merge_role(self, *, name: str, rpc_capabilities: authz.RPCCapabilities,
                             pubsub_capabilities: authz.PubsubCapabilities, **kwargs) -> bool:
        ...

    @abstractmethod
    def create_or_merge_agent_group(self,
                                    *,
                                    name: str,
                                    identities: list[authz.Identity],
                                    roles: Optional[authz.AgentRoles] = None,
                                    rpc_capabilities: Optional[authz.RPCCapabilities] = None,
                                    pubsub_capabilities: Optional[authz.PubsubCapabilities] = None,
                                    **kwargs) -> bool:
        ...

    @abstractmethod
    def remove_agents_from_group(self, name: str, identities: list[authz.Identity]):
        ...

    @abstractmethod
    def add_agents_to_group(self, name: str, identities: list[authz.Identity]):
        ...

    @abstractmethod
    def create_or_merge_agent_authz(self,
                                    *,
                                    identity: str,
                                    protected_rpcs: list[str] = None,
                                    roles: authz.AgentRoles = None,
                                    rpc_capabilities: authz.RPCCapabilities = None,
                                    pubsub_capabilities: authz.PubsubCapabilities = None,
                                    comments: str = None,
                                    **kwargs) -> bool:
        ...

    @abstractmethod
    def get_agent_capabilities(self, *, identity: str) -> dict:
        ...

    @abstractmethod
    def client_connected(self, *, credentials: Credentials):
        """
        The client_connected method is called when a client connects to the
        message bus.

        :param credentials: The credentials of the client that connected.
        :type credentials: Credentials
        """
        ...

    # def add_credential_to_role(credential: Credentials, role: str) -> None:
    #     ...

    @abstractmethod
    def create_protected_topics(self, *, topic_name_patterns: list[str]) -> bool:
        ...

    @abstractmethod
    def remove_protected_topics(self, *, topic_name_patterns: list[str]) -> bool:
        ...

    @abstractmethod
    def is_protected_topic(self, *, topic_name_pattern: str) -> bool:
        """Return True if the topic or pattern is protected, False otherwise.

        The topic_expression can be a str or regex pattern.  If the string or
        expression matches a protected topic then True is returned.  If not,
        False is returned.

        :param topic_name_pattern: The topic to check if it is protected.
        :type topic_name_pattern: str
        :return: True if the topic is protected, False otherwise.
        :rtype: bool
        """
        ...

    @abstractmethod
    def remove_agent_authorization(self, name: authz.Identity):
        ...

    @abstractmethod
    def remove_agent_group(self, name: str):
        ...

    @abstractmethod
    def remove_role(self, name: str):
        ...



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
from collections import defaultdict
from datetime import datetime

from volttron.client.known_identities import (CONTROL_CONNECTION, PLATFORM_HEALTH,
                                              PROCESS_IDENTITIES, CONTROL)
from volttron.client.vip.agent import RPC, Agent, Core
from volttron.server.decorators import service
# TODO: rmq addition
# from volttron.utils.rmq_config_params import RMQConfig
# from volttron.utils.rmq_setup import start_rabbit, RabbitMQStartError
# from volttron.services.auth.auth_service import AuthEntry, AuthFile
from volttron.types import Service
from volttron.types.service_interface import ServiceInterface
from volttron.utils import format_timestamp, set_agent_identity
from volttron.server.server_options import ServerOptions

_log = logging.getLogger(__name__)


@service
class HealthService(Agent):

    class Meta:
        identity = PLATFORM_HEALTH

    def __init__(self, options: ServerOptions, **kwargs):
        kwargs["identity"] = self.Meta.identity

        with set_agent_identity(self.Meta.identity):
            super().__init__(address=options.service_address, **kwargs)

        self._health_dict: dict = {}

    def peer_added(self, peer: str):
        """
        The `peer_added` method should be called whenever an agent is connected to the
        platform.

        :param peer: The identity of the agent connected to the platform
        """
        if peer not in self._health_dict:
            self._health_dict[peer] = dict()

        health = self._health_dict[peer]

        health["peer"] = peer
        health["service_agent"] = peer in PROCESS_IDENTITIES
        health["connected"] = format_timestamp(datetime.now())

    def peer_dropped(self, peer):
        # TODO: Should there be an option for  a db/log file for agents coming and going from the platform?
        self._health_dict[peer]["disconnected"] = format_timestamp(datetime.now())
        del self._health_dict[peer]

    @RPC.export
    def get_platform_health(self):
        """
        The `get_platform_health` retrieves all of the connected agent's health structures,
        except for the `CONTROL_CONNECTION` (vctl's known identity).  Vctl's identity is used for short
        term connections and is not relevant to the core health system.

        This function returns a dictionary in the form identity: values such as the following:

        .. code-block :: json

            {
                "listeneragent-3.3_35":
                {
                    "peer": "listeneragent-3.3_35",
                    "service_agent": False,
                    "connected": "2020-10-28T12:46:58.701119",
                    "last_heartbeat": "2020-10-28T12:47:03.709605",
                    "message": "GOOD"
                }
            }

        :return:
        """
        agents = {}
        # Ignore the connection from control in the health as it will only be around for a short while.
        if len(self._health_dict.items()) > 0:
            agents = {
                k: v
                for k, v in self._health_dict.items()
                if isinstance(v, dict) and v.get("peer") != CONTROL_CONNECTION
            }
        _log.debug(f"get_platform_health() -> {agents}")
        return agents

    def _heartbeat_updates(self, peer, sender, bus, topic, headers, message):
        """
        This method is called whenever a publish goes on the message bus from the
        heartbeat* topic.

        :param peer:
        :param sender:
        :param bus:
        :param topic:
        :param headers:
        :param message:
        :return:
        """
        # TODO: workaround for not getting notified by message bus when new peer connects
        #  should be ok for now as client side will cache and clear cache after sometime
        if sender not in self._health_dict:
            self._health_dict[sender] = dict()

        health = self._health_dict[sender]
        time_now = format_timestamp(datetime.now())
        if not health:
            health["connected"] = time_now
            health["peer"] = sender
            health["service_agent"] = sender in PROCESS_IDENTITIES

        health["last_heartbeat"] = time_now
        health["message"] = message

    @Core.receiver("onstart")
    def onstart(self, sender, **kwargs):
        self.vip.pubsub.subscribe("pubsub", "heartbeat", callback=self._heartbeat_updates)
        # pl = self.vip.rpc.call(CONTROL, "peerlist").get()
        # for peer in pl:
        #     self._health_dict[peer] =



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
import inspect
import logging
import re
import weakref
from base64 import b64decode, b64encode
from collections import defaultdict
from functools import partial

import gevent
from gevent.queue import Queue

from volttron.client.known_identities import PLATFORM_TAGGING, AUTH
from volttron.client.messaging.health import STATUS_BAD
from volttron.types.auth import AuthException
from volttron.utils import jsonapi, get_logger
from volttron.utils.scheduling import periodic
from volttron.types.message import Message

from ..decorators import annotate, annotations, dualmethod, spawn
from ..results import ResultsDictionary
from .base import SubsystemBase

__all__ = ["PubSub"]

min_compatible_version = "3.0"
max_compatible_version = ""

_log = get_logger()


def encode_peer(peer):
    if peer.startswith("\x00"):
        return peer[:1] + b64encode(peer[1:])
    return peer


def decode_peer(peer):
    if peer.startswith("\x00"):
        return peer[:1] + b64decode(peer[1:])
    return peer


class PubSub(SubsystemBase):
    """
    Pubsub subsystem concrete class implementation for ZMQ message bus.
    """

    def __init__(self, core, rpc_subsys, peerlist_subsys, owner, tag_vip_id=PLATFORM_TAGGING, tag_refresh_interval=-1):
        self.__core__ = weakref.ref(core)
        self.__rpc__ = weakref.ref(rpc_subsys)
        self.__peerlist__ = weakref.ref(peerlist_subsys)
        self.__owner__ = owner

        self._create_message_for_router = partial(Message.create_message,
                                                  peer="",
                                                  user=self.__core__().identity,
                                                  subsystem="pubsub")

        def platform_subscriptions():
            return defaultdict(subscriptions)

        def subscriptions():
            return defaultdict(set)

        # d[platform][bus][prefix] = set(callback)
        self._my_subscriptions = defaultdict(platform_subscriptions)

        # d[platform][bus][(topic_source,tag_query_condition)] = set(callback)
        self._my_tag_condition_callbacks = defaultdict(platform_subscriptions)

        # format: # d[platform][bus][prefix] = set(callback)
        # same format as _peer_subscriptions but this is updated periodically by running the
        # tag query condition and update the prefix list
        self._my_subscriptions_by_tags = defaultdict(platform_subscriptions)

        core.register("pubsub", self._handle_subsystem, self._handle_error)
        # self.vip_socket = None
        self._results = ResultsDictionary()
        self._event_queue = Queue()
        self._retry_period = 300.0
        self._processgreenlet = None
        self.tag_refresh_interval = tag_refresh_interval
        self.tag_vip_id = tag_vip_id
        self._connection = core.connection

        def setup(sender, **kwargs):
            # pylint: disable=unused-argument
            self._processgreenlet = gevent.spawn(self._process_loop)
            core.onconnected.connect(self._connected)

            #self.vip_socket = self.core().socket

            def subscribe(member):    # pylint: disable=redefined-outer-name
                for peer, bus, prefix, all_platforms, queue in annotations(member, set, "pubsub.subscriptions"):
                    # XXX: needs updated in light of onconnected signal
                    self._add_subscription("prefix", prefix, member, bus, all_platforms)
                    _log.debug("SYNC ZMQ: all_platforms {}".format(self._my_subscriptions['internal'][bus][prefix]))

                for peer, bus, tag_condition, topic_source, all_platforms, queue in annotations(
                        member, set, "pubsub.subscription_by_tags"):
                    # XXX: needs updated in light of onconnected signal
                    self.subscribe_by_tags(peer='pubsub',
                                           tag_condition=tag_condition,
                                           callback=member,
                                           topic_source=topic_source,
                                           bus=bus,
                                           all_platforms=all_platforms)

            inspect.getmembers(owner, subscribe)

        core.onsetup.connect(setup, self)
        if self.tag_refresh_interval > 0:
            core.schedule(periodic(self.tag_refresh_interval), self.refresh_tag_subscriptions)

    def _connected(self, sender, **kwargs):
        """
        Synchronize local subscriptions with PubSubService upon receiving connected signal.
        param sender: identity of sender
        type sender: str
        param kwargs: optional arguments
        type kwargs: pointer to arguments
        """
        self.synchronize()

    def _process_callback(self, sender, bus, topic, headers, message):
        """Handle incoming subscription pushes from PubSubService. It iterates over all subscriptions to find the
        subscription matching the topic and bus. It then calls the corresponding callback on finding a match.
        param sender: identity of the publisher
        type sender: str
        param bus: bus
        type bus: str
        param topic: publishing topic
        type topic: str
        param headers: header information for the incoming message
        type headers: dict
        param message: actual message
        type message: dict
        """
        peer = "pubsub"

        handled = dict()
        for platform in self._my_subscriptions:
            # _log.debug("SYNC: process callback subscriptions: {}".format(self._my_subscriptions[platform][bus]))
            buses = self._my_subscriptions[platform]
            for bus in buses:
                subscriptions = buses[bus]
                for prefix, callbacks in subscriptions.items():
                    if topic.startswith(prefix):
                        handled[prefix] = callbacks
                        for callback in callbacks:
                            callback(peer, sender, bus, topic, headers, message)
        for platform in self._my_subscriptions_by_tags:
            buses = self._my_subscriptions_by_tags[platform]
            for bus in buses:
                subscriptions = buses[bus]
                for prefix, callbacks in subscriptions.items():
                    if topic.startswith(prefix):
                        for callback in callbacks:
                            # don't call same callback function twice for the same topic
                            handled_callbacks = handled.get(prefix, set())
                            if callback not in handled_callbacks:
                                callback(peer, sender, bus, topic, headers, message)
                                handled[prefix] = callbacks

        if not handled:
            # No callbacks for topic; synchronize with sender
            self.synchronize()

    def get_topics_by_tag(self, condition):
        topics = self.rpc().call(self.tag_vip_id, "get_topics_by_tags", condition=condition).get(timeout=10)
        return topics

    @spawn
    def refresh_tag_subscriptions(self):

        def platform_subscriptions():
            return defaultdict(subscriptions)

        def subscriptions():
            return defaultdict(set)

        # format d[platform][bus][prefix] = set(callbacks)
        subscriptions_by_tag = defaultdict(platform_subscriptions)
        for platform, bus_subscriptions in self._my_tag_condition_callbacks.items():
            for bus, tag_conditions in bus_subscriptions.items():
                for (source, condition), callbacks in tag_conditions.items():
                    for prefix in self.get_topics_by_tag(condition):
                        if source:
                            prefix = source + "/" + prefix
                        subscriptions_by_tag[platform][bus][prefix] = callbacks
        self._my_subscriptions_by_tags = subscriptions_by_tag
        self.synchronize()

    def synchronize(self):
        """Synchronize local subscriptions with the PubSubService."""
        result = next(self._results)

        subscriptions_prefix_and_tag = {
            platform: {
                bus: list(subscriptions.keys())
            }
            for platform, bus_subscriptions in self._my_subscriptions.items()
            for bus, subscriptions in bus_subscriptions.items()
        }

        # extend subscriptions dict to be sent to server with the subscriptions derived from tag based subscriptions
        for platform, bus_subscriptions in self._my_subscriptions_by_tags.items():
            if not subscriptions_prefix_and_tag.get(platform):
                subscriptions_prefix_and_tag[platform] = dict()
            for bus, _subscriptions in bus_subscriptions.items():
                if not subscriptions_prefix_and_tag[platform].get(bus):
                    subscriptions_prefix_and_tag[platform][bus] = list()
                for subscription in _subscriptions.keys():
                    if subscription not in subscriptions_prefix_and_tag[platform][bus]:
                        subscriptions_prefix_and_tag[platform][bus].append(subscription)

        sync_msg = jsonapi.dumpb(dict(subscriptions=subscriptions_prefix_and_tag))
        frames = ["synchronize", "connected", sync_msg]

        message = self._create_message_for_router(msg_id=result.ident, args=frames)
        self.__core__().send_vip_message(message)

    def list(
        self,
        peer,
        prefix="",
        bus="",
        subscribed=True,
        reverse=False,
        all_platforms=False,
    ):
        """Gets list of subscriptions matching the prefix and bus for the specified peer.
        param peer: peer
        type peer: str
        param prefix: prefix of a topic
        type prefix: str
        param bus: bus
        type bus: bus
        param subscribed: subscribed or not
        type subscribed: boolean
        param reverse: reverse
        type reverse:
        :returns: List of subscriptions, i.e, list of tuples of bus, topic and
        flag to indicate if peer is a subscriber or not
        :rtype: list of tuples
        :Return Values:
        List of tuples [(topic, bus, flag to indicate if peer is a subscriber or not)]
        """
        result = next(self._results)
        list_msg = jsonapi.dumpb(
            dict(
                prefix=prefix,
                all_platforms=all_platforms,
                subscribed=subscribed,
                reverse=reverse,
                bus=bus,
            ))

        frames = ["list", list_msg]

        message = self._create_message_for_router(msg_id=result.ident, args=frames)
        self.__core__().send_vip_message(message=message)
        return result

    def _add_subscription(self, subscription_type, prefix, callback, bus="", all_platforms=False):
        # _log.debug(f"Adding subscription prefix: {prefix} allplatforms: {all_platforms}")
        if subscription_type == "prefix":
            subscription_dict = self._my_subscriptions
        elif subscription_type == "tags":
            subscription_dict = self._my_subscriptions_by_tags
        else:
            raise ValueError(f"Invalid subscription type {subscription_type}")

        if not callable(callback):
            raise ValueError("callback %r is not callable" % (callback, ))

        if not all_platforms:
            subscription_dict["internal"][bus][prefix].add(callback)
        else:
            subscription_dict["all"][bus][prefix].add(callback)
            # _log.debug("SYNC: add subscriptions: {}".format(self._my_subscriptions['internal'][bus][prefix]))

    def call_server_subscribe(self, all_platforms, bus, prefix):
        result = next(self._results)
        sub_msg = jsonapi.dumpb(dict(prefix=prefix, bus=bus, all_platforms=all_platforms))
        frames = ["subscribe", sub_msg]
        message = self._create_message_for_router(msg_id=result.ident, args=frames)
        self.__core__().connection.send_vip_message(message=message)
        #
        # self._connection.send_vip_message(message=message)
        return result

    @dualmethod
    @spawn
    def subscribe(self, peer, prefix, callback, bus="", all_platforms=False, **kwargs):
        """Subscribe to topic and register callback.

        Subscribes to topics beginning with prefix. If callback is
        supplied, it should be a function taking four arguments,
        callback(peer, sender, bus, topic, headers, message), where peer
        is the ZMQ identity of the bus owner sender is identity of the
        publishing peer, topic is the full message topic, headers is a
        case-insensitive dictionary (mapping) of message headers, and
        message is a possibly empty list of message parts
        :param peer
        :type peer str
        :param prefix topic prefix
        :type prefix str
        :param callback method to callback
        :type callback method
        :param bus message bus
        :type bus str
        :param all_platforms
        :type all_platforms boolean
        :returns: Subscribe is successful or not
        :rtype: boolean
        :Return Values:
        Success or Failure
        """
        authorized = True
        identity = self.__core__().identity
        if AUTH in self.__peerlist__().list().get():
            authorized = self.__rpc__().call("platform.auth",
                                             "check_pubsub_authorization",
                                             identity=identity,
                                             topic_pattern=prefix,
                                             access="subscribe").get()
        if authorized:
            self._add_subscription("prefix", prefix, callback, bus, all_platforms)
            return self.call_server_subscribe(all_platforms, bus, prefix)
        else:
            self.__owner__.health.set_status(STATUS_BAD, f"{identity} is not authorized to subscribe to {prefix}")
            # no harm in publishing so we don't wait till next heart beat for status update
            self.__owner__.health.publish()
            _log.error(f"{identity} is not authorized to subscribe to protected topic {prefix}")

    @dualmethod
    @spawn
    def subscribe_by_tags(self,
                          peer,
                          tag_condition,
                          callback,
                          bus="",
                          all_platforms=False,
                          topic_source="devices",
                          **kwargs):
        """Subscribe to topic based on given tags and register callback.

        Subscribes to topics that match a given combination of tags. tag_condition is a condition string using which
        tagging service can be queried for topic prefix that match the condition.
        For example - "campusRef=building1 and equip and ahu"
        If callback is supplied, it should be a function taking four arguments,
        callback(peer, sender, bus, topic, headers, message), where peer is the ZMQ identity of the bus owner sender
        is identity of the publishing peer, topic is the full message topic, headers is a case-insensitive dictionary
        (mapping) of message headers, and message is a possibly empty list of message parts
        :param peer
        :type peer
        :param tag_condition query string/condition containing tags that need be matched
        :type tag_condition str
        :param callback method to callback
        :type callback method
        :param bus message bus
        :type bus str
        :param all_platforms
        :type all_platforms boolean
        :param topic_source message bus topic source. Will get added to beginning of each of the topics that matches the
         given tag condition. defaults to "devices"
        :type topic_source str
        :returns: success_list, failure_list
        :rtype: list, list

        :Return Values:
        [List of matched prefix successfully subscribed to], [list of matched prefix we couldn't subscribe to]
        """
        if all_platforms:
            platform = "all"
        else:
            platform = "internal"

        # Query tagging service to topic prefix that match the given tag search condition
        topic_prefixes = self.get_topics_by_tag(tag_condition)
        if not topic_prefixes:
            raise ValueError(f"No topics match given tag condition {tag_condition}")
        success_list = []
        failure_list = []
        for prefix in topic_prefixes:
            if topic_source:
                prefix = topic_source + "/" + prefix
            self._add_subscription("tags", prefix, callback, bus, all_platforms)
            if self.call_server_subscribe(all_platforms, bus, prefix):
                success_list.append(prefix)
            else:
                failure_list.append(prefix)

        if success_list:
            # even if there was one successful subscription save tag_condition for periodic updates
            self._my_tag_condition_callbacks[platform][bus][(topic_source, tag_condition)].add(callback)
        return success_list, failure_list

    @subscribe.classmethod
    def subscribe(cls, peer, prefix, bus="", all_platforms=False, persistent_queue=None):

        def decorate(method):
            annotate(
                method,
                set,
                "pubsub.subscriptions",
                (peer, bus, prefix, all_platforms, persistent_queue),
            )
            return method

        return decorate

    @subscribe_by_tags.classmethod
    def subscribe_by_tags(cls,
                          peer,
                          tag_condition,
                          bus="",
                          all_platforms=False,
                          persistent_queue=None,
                          topic_source="devices"):

        def decorate(method):
            annotate(
                method,
                set,
                "pubsub.subscription_by_tags",
                (peer, bus, tag_condition, topic_source, all_platforms, persistent_queue),
            )
            return method

        return decorate

    def _drop_subscription(self, subscription_type, prefix, callback, bus="", platform="internal"):
        """
        Drop the subscription for the specified prefix, callback and bus.
        param prefix: prefix to be removed
        type prefix: str
        param callback: callback method
        type callback: method
        param bus: bus
        type bus: bus
        return: list of topics/prefixes
        :rtype: list
        :Return Values:
        List of prefixes
        """
        if subscription_type == "prefix":
            subscription_dict = self._my_subscriptions
        elif subscription_type == "tags":
            subscription_dict = self._my_subscriptions_by_tags
        else:
            raise ValueError(f"Invalid subscription type {subscription_type}")

        topics = []
        bus_subscriptions = dict()
        if prefix is None:
            if callback is None:
                if len(subscription_dict) and platform in subscription_dict:
                    bus_subscriptions = subscription_dict[platform]
                    if bus in bus_subscriptions:
                        topics.extend(bus_subscriptions[bus].keys())
                if not len(topics):
                    return []
            else:
                if platform in subscription_dict:
                    bus_subscriptions = subscription_dict[platform]
                if bus in bus_subscriptions:
                    subscriptions = bus_subscriptions[bus]
                    remove = []
                    for topic, callbacks in subscriptions.items():
                        try:
                            callbacks.remove(callback)
                        except KeyError:
                            pass
                        else:
                            topics.append(topic)
                        if not callbacks:
                            remove.append(topic)
                    for topic in remove:
                        del subscriptions[topic]
                    if not subscriptions:
                        del bus_subscriptions[bus]
                    if not bus_subscriptions:
                        del subscription_dict[platform]
            if not topics:
                raise KeyError("no such subscription")
        else:
            _log.debug(f"BEFORE: {subscription_dict}")
            if platform in subscription_dict:
                bus_subscriptions = subscription_dict[platform]
                if bus in bus_subscriptions:
                    _log.debug(f"BUS: {bus}")
                    subscriptions = bus_subscriptions[bus]
                    _log.debug(f"subscriptions: {subscriptions}")
                    if callback is None:
                        try:
                            del subscriptions[prefix]
                        except KeyError:
                            return []
                    else:
                        try:
                            callbacks = subscriptions[prefix]
                            _log.debug(f"callbacks: {callbacks}")
                        except KeyError:
                            return []
                        try:
                            callbacks.remove(callback)
                        except KeyError as e:
                            _log.debug(f"KeyError: {e}")
                            pass
                        if not callbacks:
                            try:
                                del subscriptions[prefix]
                                _log.debug(f"subscriptions: {subscriptions}")
                            except KeyError:
                                return []
                    topics = [prefix]
                    if not subscriptions:
                        del bus_subscriptions[bus]
                    if not bus_subscriptions:
                        del subscription_dict[platform]
        _log.debug(f"AFTER: {subscription_dict}")
        return topics

    def call_server_unsubscribe(self, bus, platform, subscriptions, topics):
        result = next(self._results)
        subscriptions[platform] = dict(prefix=topics, bus=bus)
        unsub_msg = jsonapi.dumpb(subscriptions)
        frames = ["unsubscribe", unsub_msg]

        message = self._create_message_for_router(msg_id=result.ident, args=frames)
        self._connection.send_vip_message(message=message)
        return result

    def unsubscribe(self, peer, prefix, callback, bus="", all_platforms=False, **kwargs):
        """Unsubscribe and remove callback(s).

        Remove all handlers matching the given info - peer, callback and bus, which was used earlier to subscribe as
        well. If all handlers for a topic prefix are removed, the topic is also unsubscribed.
        param peer: peer
        type peer: str
        param prefix: prefix that needs to be unsubscribed
        type prefix: str
        param callback: callback method
        type callback: method
        param bus: bus
        type bus: bus
        return: success or not
        :rtype: boolean
        :Return Values:
        success or not
        """
        subscriptions = dict()

        if not all_platforms:
            platform = "internal"
        else:
            platform = "all"

        topics = self._drop_subscription("prefix", prefix, callback, bus, platform)
        return self.call_server_unsubscribe(bus, platform, subscriptions, topics)

    @spawn
    def unsubscribe_by_tags(self,
                            peer,
                            tag_condition,
                            callback,
                            bus="",
                            all_platforms=False,
                            topic_source="devices",
                            **kwargs):
        """Unsubscribe to topic based on given tags and register callback.

        Subscribes to topics that match a given combination of tags. tag_condition is a condition string using which
        tagging service can be queried for topic prefix that match the condition.
        For example - "campusRef=building1 and equip and ahu"
        If callback is supplied, it should be a function taking four arguments,
        callback(peer, sender, bus, topic, headers, message), where peer is the ZMQ identity of the bus owner sender
        is identity of the publishing peer, topic is the full message topic, headers is a case-insensitive dictionary
        (mapping) of message headers, and message is a possibly empty list of message parts
        :param peer
        :type peer
        :param tag_condition query string/condition containing tags that need be matched
        :type tag_condition str
        :param callback method to callback
        :type callback method
        :param bus message bus
        :type bus str
        :param all_platforms
        :type all_platforms boolean
        :param topic_source message bus topic source. Will get added to beginning of each of the topics that matches the
         given tag condition. defaults to "devices"
        :type topic_source str
        :returns: success_list, failure_list
        :rtype: list, list

        :Return Values:
        [List of matched prefix successfully unsubscribed], [list of matched prefix we couldn't unsubscribe]
        """
        subscriptions = dict()

        if all_platforms:
            platform = "all"
        else:
            platform = "internal"

        if not tag_condition:
            raise KeyError("tag_condition is mandatory")

        # Query tagging service to topic prefix that match the given tag search condition
        topic_prefixes = self.get_topics_by_tag(tag_condition)

        if not topic_prefixes:
            raise KeyError(f"Not topics match given tag condition {tag_condition}")
        success_list = []
        failure_list = []
        for prefix in topic_prefixes:
            if topic_source:
                prefix = topic_source + "/" + prefix
            topics = self._drop_subscription("tags", prefix, callback, bus, platform)
            if self.call_server_unsubscribe(bus, platform, subscriptions, topics):
                success_list.extend(topics)
            else:
                failure_list.extend(topics)

        if not failure_list:
            # remove tag_condition only if there were no failures. if there are failures leave tag condition in place
            # so that user could call unsubscribe again
            remove = []
            for platform, bus_subscriptions in self._my_tag_condition_callbacks.items():
                for bus, tag_subscriptions in bus_subscriptions.items():
                    for t, callbacks in tag_subscriptions.items():
                        if t == (topic_source, tag_condition):
                            if callback:
                                try:
                                    callbacks.remove(callback)
                                except KeyError:
                                    pass
                            # if passed callback is none or if no callbacks left after last callbacks.remove()
                            if not callback or not callbacks:
                                remove.append((platform, bus, (topic_source, tag_condition)))

            for platform, bus, condition in remove:
                subscriptions = self._my_tag_condition_callbacks[platform][bus]
                subscriptions.pop(condition)
        return success_list, failure_list

    def publish(self, peer: str, topic: str, headers=None, message=None, bus="", **kwargs):
        """
        Publish a message to a given topic via a peer.

        Publish headers and message to all subscribers of topic on bus.
        If peer is None, use self. Adds volttron platform version
        compatibility information to header as variables
        min_compatible_version and max_compatible version
        param peer: peer
        type peer: str
        param topic: topic to publish to
        type topic: str
        param headers: header info for the message
        type headers: None or dict
        param message: actual message
        type message: None or any
        param bus: bus
        type bus: str
        return: async result - contains Number of subscribers the message was sent to.
        :rtype: AsyncResult
        """
        if headers is None:
            headers = {}
        headers["min_compatible_version"] = min_compatible_version
        headers["max_compatible_version"] = max_compatible_version

        if peer is None:
            peer = "pubsub"
        authorized = True
        identity = self.__core__().identity
        if AUTH in self.__peerlist__().list().get():
            authorized = self.__rpc__().call("platform.auth",
                                             "check_pubsub_authorization",
                                             identity=identity,
                                             topic_pattern=topic,
                                             access="publish").get()
        result = next(self._results)
        if authorized:
            args = ["publish", topic, dict(bus=bus, headers=headers, message=message)]
            message = self._create_message_for_router(msg_id=result.ident, args=args)
            _log.debug(f"sending pubsub message created for router is: {message}")
            self.__core__().connection.send_vip_message(message=message)
        else:
            self.__owner__.health.set_status(STATUS_BAD, f"{identity} is not authorized to subscribe to {topic}")
            self.__owner__.health.publish()
            _log.error(f"{identity} is not authorized to subscribe to protected topic {topic}")

        return result

    def publish_by_tags(self,
                        peer: str,
                        tag_condition: str,
                        headers=None,
                        message=None,
                        bus="",
                        max_publish_count=1,
                        topic_source="devices",
                        **kwargs):
        """Publish a message to a topic that matches the give tag_condition via a peer. If tag_condition resolves to
        more than one topic then throw an error if publish_multiple is False. Publish to multiple matching topics if
        publish_multiple parameter is True

        Publish headers and message to all subscribers of topic on bus.
        If peer is None, use self. Adds volttron platform version
        compatibility information to header as variables
        min_compatible_version and max_compatible version
        :param peer: peer
        :type peer: str
        :param tag_condition: tag_condition to find topics to publish to
        :type tag_condition: str
        :param headers: header info for the message
        :type headers: None or dict
        :param message: actual message
        :type message: None or any
        :param bus: bus
        :type bus: str
        :param max_publish_count: maximum number of publish that can be done. By default expects the tag_condition
         to match a single topic.
        :type max_publish_count: int
        :param topic_source message bus topic source. Will get added to beginning of each of the topics that matches the
         given tag condition. defaults to "devices"
        :type topic_source str

        """
        if not tag_condition:
            raise KeyError("tag_condition is mandatory")
        number_of_subscribers = 0

        topic_prefixes = []
        # Query tagging service to topic prefix that match the given tag search condition
        topic_prefixes = self.get_topics_by_tag(tag_condition)
        if not topic_prefixes:
            raise ValueError(f"Not topics match given tag condition {tag_condition}")
        count = len(topic_prefixes)
        if count > max_publish_count:
            raise ValueError(f"tag condition {tag_condition} matched {count} topics "
                             f"but max_publish_count is set to {max_publish_count}")
        for topic in topic_prefixes:
            if topic_source:
                topic = topic_source + "/" + topic
            self.publish(peer, topic, headers, message, bus)

    def _handle_subsystem(self, message):
        """Handler for incoming messages
        param message: VIP message from PubSubService
        type message: dict
        """
        _log.debug(f"Putting message in event queue for {self.__core__().identity} {message}")
        self._event_queue.put(message)

    @spawn
    def _process_incoming_message(self, message):
        """Process incoming messages
        param message: VIP message from PubSubService
        type message: dict
        """
        op = message.args[0]

        _log.debug(f"Processing {self.__core__().identity}: op: {op}, message: {message}")
        response = None
        if op == "request_response":
            result = None
            try:
                result = self._results.pop(message.id)
            except KeyError:
                pass
            _log.debug(f"Result is: {result}")
            response = message.args[1]

            if result:
                result.set(response)

        elif op == "publish":
            try:
                topic = message.args[1]
                msg = message.args[2]
            except IndexError:
                return
            try:
                headers = msg["headers"]
                message = msg["message"]
                sender = msg["sender"]
                bus = msg["bus"]
            except KeyError as exc:
                _log.error("Missing keys in pubsub message: {}".format(exc))
            else:
                response = message
                self._process_callback(sender, bus, topic, headers, message)

        elif op == "list_response":
            result = None
            try:
                result = self._results.pop(message.id)
                response = message.args[1]
                if result:
                    result.set(response)
            except KeyError:
                pass
        else:
            _log.error("Unknown operation ({})".format(op))

        if response is not None:
            _log.debug(f"Processed {op} response was {response}")

    def _process_loop(self):
        """Incoming message processing loop"""
        for msg in self._event_queue:
            _log.debug(f"Handling pubsub message: {msg}")
            self._process_incoming_message(msg)

    def _handle_error(self, sender, message, error, **kwargs):
        """Error handler. If UnknownSubsystem error is received, it implies that agent is connected to platform that has
        OLD pubsub implementation. So messages are resent using RPC method.
        :param message: Error message
        :type message: dict
        :param error: indicates error type
        :type error: error class
        :param **kwargs: variable arguments
        :type **kwargs: dict
        """
        try:
            result = self._results.pop(message.id)
        except KeyError:
            return
        result.set_exception(error)
